"""Base class for euporie kernels."""

from __future__ import annotations

import asyncio
import concurrent
import logging
import threading
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, NamedTuple, TypedDict, overload

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine
    from typing import Any, Literal, Protocol, TypeVar, Unpack

    from euporie.core.tabs.kernel import KernelTab

    class KernelFactory(Protocol):
        """Type for kernel factory functions."""

        def __call__(
            self,
            kernel_tab: KernelTab,
            default_callbacks: MsgCallbacks | None = None,
            allow_stdin: bool = False,
            **kwargs: Any,
        ) -> BaseKernel:
            """Signature for creating a new Kernel instance."""

    T = TypeVar("T")

log = logging.getLogger(__name__)

_THREAD: list[threading.Thread | None] = [None]
_LOOP: list[asyncio.AbstractEventLoop | None] = [None]


class KernelInfo(NamedTuple):
    """Named tuple representing a launchable kernel."""

    name: str
    display_name: str
    type: type[BaseKernel]
    kind: Literal["new", "existing"]
    factory: KernelFactory


class MsgCallbacks(TypedDict, total=False):
    """Typed dictionary for named message callbacks."""

    get_input: Callable[[str, bool], None] | None
    set_execution_count: Callable[[int], None] | None
    add_output: Callable[[dict[str, Any], bool], None] | None
    add_input: Callable[[dict[str, Any], bool], None] | None
    clear_output: Callable[[bool], None] | None
    done: Callable[[dict[str, Any]], None] | None
    set_metadata: Callable[[tuple[str, ...], Any], None] | None
    set_status: Callable[[str], None] | None
    set_kernel_info: Callable[[dict[str, Any]], None] | None
    completeness_status: Callable[[dict[str, Any]], None] | None
    dead: Callable[[], None] | None
    # Payloads
    page: Callable[[list[dict], int], None] | None
    set_next_input: Callable[[str, bool], None] | None
    edit_magic: Callable[[str, int], None] | None
    ask_exit: Callable[[bool], None] | None


def get_loop() -> asyncio.AbstractEventLoop:
    """Create or return the conversion IO loop.

    The loop will be running on a separate thread.
    """
    if _LOOP[0] is None:
        loop = asyncio.new_event_loop()
        _LOOP[0] = loop
        thread = threading.Thread(
            target=loop.run_forever, name="EuporieKernelLoop", daemon=True
        )
        thread.start()
        _THREAD[0] = thread
    assert _LOOP[0] is not None
    # Check we are not already in the conversion event loop
    try:
        running_loop = asyncio.get_running_loop()
    except RuntimeError:
        running_loop = None
    if _LOOP[0] is running_loop:
        raise NotImplementedError("Cannot nest event loop access")
    return _LOOP[0]


class BaseKernel(ABC):
    """Abstract base class for euporie kernels."""

    @classmethod
    def variants(cls) -> list[KernelInfo]:
        """Return a list of parameterized variants of this kernel."""
        return []

    def __init__(
        self,
        kernel_tab: KernelTab,
        default_callbacks: MsgCallbacks | None = None,
        allow_stdin: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize the kernel.

        Args:
            kernel_tab: The notebook this kernel belongs to
            allow_stdin: Whether the kernel is allowed to request input
            default_callbacks: The default callbacks to use on receipt of a message
            **kwargs: Additional keyword arguments passed to parent classes
        """
        self.loop = get_loop()
        self.kernel_tab = kernel_tab
        self.allow_stdin = allow_stdin
        self._status = "stopped"
        self.error: Exception | None = None
        self.dead = False
        self.status_change_event = asyncio.Event()
        self.coros: dict[str, concurrent.futures.Future] = {}
        self.msg_id_callbacks: dict[str, MsgCallbacks] = {}
        self.threaded = False

        self.default_callbacks = MsgCallbacks(
            {
                "get_input": None,
                "set_execution_count": None,
                "add_output": None,
                "clear_output": None,
                "done": None,
                "set_metadata": None,
                "set_status": None,
            }
        )
        if default_callbacks is not None:
            self.default_callbacks.update(default_callbacks)

    @property
    @abstractmethod
    def spec(self) -> dict[str, str]:
        """The kernelspec metadata for the current kernel instance."""

    @overload
    def _aodo(
        self,
        coro: Coroutine[Any, Any, T],
        wait: Literal[True] = True,
        callback: Callable | None = None,
        timeout: int | float | None = None,
        single: bool = False,
    ) -> T | None: ...
    @overload
    def _aodo(
        self,
        coro: Coroutine[Any, Any, T],
        wait: Literal[False] = False,
        callback: Callable | None = None,
        timeout: int | float | None = None,
        single: bool = False,
    ) -> concurrent.futures.Future: ...
    @overload
    def _aodo(
        self,
        coro: Coroutine[Any, Any, T],
        wait: bool = False,
        callback: Callable | None = None,
        timeout: int | float | None = None,
        single: bool = False,
    ) -> T | None | concurrent.futures.Future: ...
    def _aodo(self, coro, wait=False, callback=None, timeout=None, single=False):
        """Schedule a coroutine in the kernel's event loop.

        Optionally waits for the results (blocking the main thread). Optionally
        schedules a callback to run when the coroutine has completed or timed out.

        Args:
            coro: The coroutine to run
            wait: If :py:const:`True`, block until the kernel has started
            callback: A function to run when the coroutine completes. The result from
                the coroutine will be passed as an argument
            timeout: The number of seconds to allow the coroutine to run if waiting
            single: If :py:const:`True`, any futures for previous instances of the
                coroutine will be cancelled

        Returns:
            The result of the coroutine

        """
        future = asyncio.run_coroutine_threadsafe(coro, self.loop)

        # Cancel previous future instances if required
        if single and self.coros.get(coro.__name__):
            self.coros[coro.__name__].cancel()
        self.coros[coro.__name__] = future

        if wait:
            result = None
            try:
                result = future.result(timeout)
            except concurrent.futures.TimeoutError:
                log.error("Operation '%s' timed out", coro)
                future.cancel()
                return result
            finally:
                if callable(callback):
                    callback(result)
            return result
        else:
            if callable(callback):
                future.add_done_callback(lambda f: callback(f.result()))
            return future

    @property
    def status(self) -> str:
        """Retrieve the current kernel status.

        Returns:
            The kernel status

        """
        return self._status

    @status.setter
    def status(self, value: str) -> None:
        """Set the kernel status."""
        self.status_change_event.set()
        self._status = value
        self.status_change_event.clear()

    def wait_for_status(self, status: str = "idle") -> None:
        """Block until the kernel reaches a given status value."""
        if self.status != status:

            async def _wait() -> None:
                while self.status != status:
                    await asyncio.wait_for(
                        self.status_change_event.wait(), timeout=None
                    )

            self._aodo(_wait(), wait=True)

    def start(
        self, cb: Callable | None = None, wait: bool = False, timeout: int = 10
    ) -> None:
        """Start the kernel.

        Args:
            cb: An optional callback to run after the kernel has started
            wait: If :py:const:`True`, block until the kernel has started
            timeout: How long to wait until failure is assumed

        """
        self._aodo(
            self.start_async(),
            timeout=timeout,
            wait=wait,
            callback=cb,
        )

    @abstractmethod
    async def start_async(self) -> None:
        """Start the kernel."""

    def stop(self, cb: Callable | None = None, wait: bool = False) -> None:
        """Stop the kernel."""
        log.debug("Stopping kernel %s (wait=%s)", self.id, wait)
        if not wait:
            self.interrupt()
        self._aodo(self.stop_async(), callback=cb, wait=wait)

    async def stop_async(self, cb: Callable | None = None) -> None:
        """Stop the kernel asynchronously."""
        self.stop(cb)

    def run(
        self,
        source: str,
        wait: bool = False,
        callback: Callable[..., None] | None = None,
        **callbacks: Callable[..., Any],
    ) -> None:
        """Execute code in the kernel."""
        self._aodo(
            self.run_async(source, **callbacks),
            wait=wait,
            callback=callback,
        )

    @abstractmethod
    async def run_async(
        self, source: str, **local_callbacks: Unpack[MsgCallbacks]
    ) -> None:
        """Execute code in the kernel asynchronously."""

    def complete(self, source: str, cursor_pos: int) -> list[dict]:
        """Request code completions from the kernel.

        Args:
            source: The code string to retrieve completions for
            cursor_pos: The position of the cursor in the code string

        Returns:
            A list of dictionaries defining completion entries. The dictionaries
            contain ``text`` (the completion text), ``start_position`` (the stating
            position of the completion text), and optionally ``display_meta``
            (a string containing additional data about the completion type)

        """
        return (
            self._aodo(
                self.complete_async(source, cursor_pos),
                wait=True,
                single=True,
            )
            or []
        )

    @abstractmethod
    async def complete_async(self, source: str, cursor_pos: int) -> list[dict]:
        """Get code completions asynchronously."""

    def history(
        self, pattern: str = "", n: int = 1, hist_access_type: str = "search"
    ) -> list[tuple[int, int, str]] | None:
        """Retrieve history from the kernel.

        Args:
            pattern: The pattern to search for
            n: the number of history items to return
            hist_access_type: How to access the history ('range', 'tail' or 'search')

        Returns:
            A list of history items, consisting of tuples (session, line_number, input)

        """
        return self._aodo(
            self.history_async(pattern, n, hist_access_type),
            wait=True,
            single=True,
        )

    async def history_async(
        self,
        pattern: str = "",
        n: int = 1,
        hist_access_type: str = "search",
        timeout: int = 1,
    ) -> list[tuple[int, int, str]] | None:
        """Retrieve history from the kernel asynchronously."""
        return []

    def inspect(
        self,
        source: str,
        cursor_pos: int,
        detail_level: int = 0,
        timeout: int = 2,
        callback: Callable[[dict[str, Any]], None] | None = None,
    ) -> dict[str, Any]:
        """Request code inspection from the kernel.

        Args:
            source: The code string to retrieve completions for
            cursor_pos: The position of the cursor in the code string
            detail_level: Level of detail for the inspection (0-2)
            timeout: Number of seconds to wait for inspection results
            callback: A function to run when the inspection result arrives. The result
                is passed as an argument.

        Returns:
            A string containing useful information about the code at the current cursor
            position

        """
        return (
            self._aodo(
                self.inspect_async(source, cursor_pos, detail_level),
                wait=True,
                callback=callback,
                single=True,
            )
            or {}
        )

    @abstractmethod
    async def inspect_async(
        self,
        source: str,
        cursor_pos: int,
        detail_level: int = 0,
        timeout: int = 2,
    ) -> dict[str, Any]:
        """Get code inspection/documentation asynchronously."""

    def is_complete(
        self,
        source: str,
        timeout: int | float = 0.1,
        callback: Callable[[dict[str, Any]], None] | None = None,
    ) -> dict[str, Any]:
        """Request code completeness status from the kernel.

        Args:
            source: The code string to check the completeness status of
            timeout: How long to wait for a kernel response
            wait: Whether to wait for the response
            callback: A function to run when the inspection result arrives. The result
                is passed as an argument.

        Returns:
            A string describing the completeness status

        """
        return (
            self._aodo(
                self.is_complete_async(source, timeout), wait=True, callback=callback
            )
            or {}
        )

    @abstractmethod
    async def is_complete_async(
        self,
        source: str,
        timeout: int | float = 0.1,
    ) -> dict[str, Any]:
        """Check if code is complete asynchronously."""

    @abstractmethod
    def input(self, text: str) -> None:
        """Send input to the kernel."""

    @abstractmethod
    def interrupt(self) -> None:
        """Interrupt the kernel."""

    def restart(self, wait: bool = False, cb: Callable | None = None) -> None:
        """Restart the current kernel."""
        self._aodo(
            self.restart_async(),
            wait=wait,
            callback=cb,
        )

    @abstractmethod
    async def restart_async(self) -> None:
        """Restart the kernel asynchronously."""

    def shutdown(self, wait: bool = False, cb: Callable | None = None) -> None:
        """Shutdown the kernel.

        This is intended to be run when the notebook is closed: the
        :py:class:`~euporie.core.kernel.base.BaseKernel` cannot be restarted after this.

        Args:
            wait: Whether to block until shutdown completes
            cb: Callback run after shutdown completes

        """
        self._aodo(self.shutdown_async(), wait=wait, callback=cb)

    @abstractmethod
    async def shutdown_async(self) -> None:
        """Shutdown the kernel asynchronously."""

    @property
    @abstractmethod
    def missing(self) -> bool:
        """Return whether the kernel is missing."""

    @property
    @abstractmethod
    def id(self) -> str | None:
        """Return the kernel ID."""

    @abstractmethod
    def info(
        self,
        set_kernel_info: Callable[[dict[str, Any]], None] | None = None,
        set_status: Callable[[str], None] | None = None,
    ) -> None:
        """Request information about the kernel."""

    def kc_comm(self, comm_id: str, data: dict[str, Any]) -> str:
        """By default kernels do not implement COMM communication."""
        log.warning("The %s kernel does not implement COMMs", self.__class__.__name__)
        return ""  # TODO - raise NotImplementedError

    def comm_info(self, target_name: str | None = None) -> None:
        """Request information about the current comms.

        Does nothing by default.
        """


class NoKernel(BaseKernel):
    """A `None` kernel."""

    @property
    def spec(self) -> dict[str, str]:
        """The kernelspec metadata for the current kernel instance."""
        raise NotImplementedError()

    async def start_async(self) -> None:
        """Start the kernel."""
        raise NotImplementedError()

    async def run_async(
        self, source: str, **local_callbacks: Unpack[MsgCallbacks]
    ) -> None:
        """Execute code in the kernel asynchronously."""
        raise NotImplementedError()

    async def is_complete_async(
        self,
        source: str,
        timeout: int | float = 0.1,
    ) -> dict[str, Any]:
        """Check if code is complete asynchronously."""
        raise NotImplementedError()

    async def complete_async(self, source: str, cursor_pos: int) -> list[dict]:
        """Get code completions asynchronously."""
        raise NotImplementedError()

    async def inspect_async(
        self,
        source: str,
        cursor_pos: int,
        detail_level: int = 0,
        timeout: int = 2,
    ) -> dict[str, Any]:
        """Get code inspection/documentation asynchronously."""
        raise NotImplementedError()

    def input(self, text: str) -> None:
        """Send input to the kernel."""
        # Do nothing

    def interrupt(self) -> None:
        """Interrupt the kernel."""
        # Do nothing

    async def restart_async(self) -> None:
        """Restart the kernel asynchronously."""
        # Do nothing

    async def shutdown_async(self) -> None:
        """Shutdown the kernel asynchronously."""
        # Do nothing

    @property
    def missing(self) -> bool:
        """Return whether the kernel is missing."""
        raise NotImplementedError()

    @property
    def id(self) -> str | None:
        """Return the kernel ID."""
        raise NotImplementedError()

    def info(
        self,
        set_kernel_info: Callable[[dict[str, Any]], None] | None = None,
        set_status: Callable[[str], None] | None = None,
    ) -> None:
        """Request information about the kernel."""
        raise NotImplementedError()



"""Contain the main class for a notebook file."""

from __future__ import annotations

import logging
from collections import deque
from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING, ClassVar

import nbformat
from prompt_toolkit.clipboard.base import ClipboardData
from prompt_toolkit.filters import (
    Condition,
)
from prompt_toolkit.layout.containers import ConditionalContainer, VSplit
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.mouse_events import MouseEventType

from euporie.core.commands import get_cmd
from euporie.core.filters import (
    insert_mode,
    multiple_cells_selected,
    replace_mode,
)
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.layout.cache import CachedContainer
from euporie.core.layout.decor import Line, Pattern
from euporie.core.layout.scroll import ScrollingContainer
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.style import KERNEL_STATUS_REPR
from euporie.core.tabs.notebook import BaseNotebook
from euporie.core.widgets.cell import Cell

if TYPE_CHECKING:
    from collections.abc import MutableSequence, Sequence
    from pathlib import Path
    from typing import Any

    from prompt_toolkit.formatted_text.base import StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.mouse_events import MouseEvent

    from euporie.core.app.app import BaseApp
    from euporie.core.bars.status import StatusBarFields
    from euporie.core.comm.base import Comm
    from euporie.core.kernel.base import BaseKernel

log = logging.getLogger(__name__)


class Notebook(BaseNotebook):
    """Interactive notebooks.

    A tab which allows running and editing a notebook.
    """

    name = "Notebook Editor"
    weight = 3
    mime_types: ClassVar[set[str]] = {"application/x-ipynb+json"}
    file_extensions: ClassVar[dict[str, None]] = {".ipynb": None}
    try:
        from jupytext.formats import NOTEBOOK_EXTENSIONS
    except ModuleNotFoundError:
        pass
    else:
        file_extensions.update(dict.fromkeys(NOTEBOOK_EXTENSIONS))

    allow_stdin = True
    bg_init = True

    def __init__(
        self,
        app: BaseApp,
        path: Path | None = None,
        kernel: BaseKernel | None = None,
        comms: dict[str, Comm] | None = None,
        use_kernel_history: bool = True,
        json: dict[str, Any] | None = None,
    ) -> None:
        """Load a editable notebook."""
        super().__init__(
            app=app,
            path=path,
            kernel=kernel,
            comms=comms,
            use_kernel_history=use_kernel_history,
            json=json,
        )

        self.edit_mode = False
        self.in_edit_mode = Condition(self.check_edit_mode)
        self.multiple_cells_selected = multiple_cells_selected
        self.clipboard: list[Cell] = []
        self.undo_buffer: deque[tuple[int, list[Cell]]] = deque(maxlen=10)
        self.default_callbacks["set_next_input"] = self.set_next_input

    # Tab stuff

    def _statusbar_kernel_handler(self, event: MouseEvent) -> NotImplementedOrNone:
        """Event handler for kernel name field in statusbar."""
        if event.event_type == MouseEventType.MOUSE_UP:
            get_cmd("change-kernel").run()
            return None
        else:
            return NotImplemented

    def __pt_status__(self) -> StatusBarFields | None:
        """Generate the formatted text for the statusbar."""
        if self.loaded:
            rendered = self.page.pre_rendered

            def _kernel_name() -> StyleAndTextTuples:
                ft: StyleAndTextTuples = [
                    (
                        "",
                        self.kernel_display_name or "No Kernel",
                        self._statusbar_kernel_handler,
                    )
                ]
                return ft

            return (
                [
                    self.mode(),
                    f"Cell {self.page.selected_slice.start + 1}",
                    f"Rendering… ({rendered:.0%})" if rendered and rendered < 1 else "",
                    "Saving…" if self.saving else "",
                ],
                [
                    _kernel_name,
                    KERNEL_STATUS_REPR[self.kernel.status] if self.kernel else ".",
                ],
            )
        else:
            return ([], [])

    # Notebook stuff

    def post_init_kernel(self) -> None:
        """Start the kernel after if has been loaded."""
        # Load container
        super().post_init_kernel()

        # Start kernel
        if self.kernel._status == "stopped":
            self.kernel.start(cb=self.kernel_started, wait=False)

    @property
    def selected_indices(self) -> list[int]:
        """Return a list of the currently selected cell indices."""
        return self.page.selected_indices

    def kernel_started(self, result: dict[str, Any] | None = None) -> None:
        """Run when the kernel has started."""
        super().kernel_started(result)

        if self.kernel.status == "idle" and self.app.config.run:
            self.run_all(wait=False)

    def report_kernel_error(self, error: Exception | None) -> None:
        """Report a kernel error to the user."""
        self.app.dialogs["error"].show(exception=error, when="starting the kernel")

    def load_container(self) -> AnyContainer:
        """Trigger loading of the main notebook container."""

        async def _load() -> None:
            # Load notebook file
            self.load()
            # Load an focus container
            prev = self.container
            self.container = self._load_container()
            self.loaded = True
            # Update the focus if the old container had focus
            if self.app.layout.has_focus(prev):
                self.focus()

        self.app.create_background_task(_load())

        return self.container

    def _load_container(self) -> AnyContainer:
        """Actually load the main notebook container."""
        self.page = ScrollingContainer(
            self.rendered_cells, width=self.app.config.max_notebook_width
        )

        expand = Condition(lambda: self.app.config.expand)

        pattern = Pattern(
            partial(getattr, self.app.config, "background_character"),
            partial(getattr, self.app.config, "background_pattern"),
        )

        return VSplit(
            [
                ConditionalContainer(
                    VSplit(
                        [
                            pattern,
                            Line(
                                char="▋",  # 5/8
                                width=1,
                                collapse=True,
                                style="class:drop-shadow,outer reverse",
                            ),
                            Line(
                                char="▎",  # 2/8
                                width=1,
                                collapse=True,
                                style="class:drop-shadow,inner",
                            ),
                        ]
                    ),
                    filter=~expand,
                ),
                self.page,
                ConditionalContainer(
                    VSplit(
                        [
                            Line(
                                char="▊",  # 6/8
                                width=1,
                                collapse=True,
                                style="class:drop-shadow,inner reverse",
                            ),
                            Line(
                                char="▍",  # 3/8
                                width=1,
                                collapse=True,
                                style="class:drop-shadow,outer",
                            ),
                            pattern,
                        ]
                    ),
                    filter=~expand,
                ),
                ConditionalContainer(
                    MarginContainer(ScrollbarMargin(), target=self.page),
                    filter=Condition(lambda: self.app.config.show_scroll_bar),
                ),
            ],
            width=Dimension(weight=1),
            height=Dimension(min=1, weight=2),
            key_bindings=load_registered_bindings(
                "euporie.core.tabs.base.Tab",
                "euporie.notebook.tabs.notebook.Notebook",
                config=self.app.config,
            ),
        )

    @property
    def cell(self) -> Cell:
        """Return the currently selected `Cell` in this `Notebook`."""
        if isinstance(cell := self.page.get_child().content, Cell):
            return cell
        return Cell(0, {}, self)

    # Editing specific stuff

    @property
    def cells(self) -> Sequence[Cell]:
        """Return the currently selected `Cells` in this `Notebook`."""
        return [
            child.content
            for child in self.page.all_children()[self.page.selected_slice]
            if isinstance(child, CachedContainer) and isinstance(child.content, Cell)
        ]

    def check_edit_mode(self) -> bool:
        """Determine if the notebook is (or should be) in edit mode."""
        if self.cell.input_box.has_focus():
            self.edit_mode = True
        return self.edit_mode

    def enter_edit_mode(self) -> None:
        """Enter cell edit mode."""
        # Signal the page to scroll so the cursor is visible on next render
        self.page.scroll_to_cursor = True
        self.edit_mode = True
        # Only one cell can be selected in edit mode
        self.select(self.cell.index)
        self.scroll_to(self.cell.index)

    def exit_edit_mode(self) -> None:
        """Leave cell edit mode."""
        self.edit_mode = False
        # Focus the first selected cell
        self.rendered_cells()[self.page.selected_slice.start].focus(
            self.cell.index, scroll=False
        )

    def mode(self) -> str:
        """Return a symbol representing the current mode.

        * ``^``: Notebook mode
        * ``>``: Navigation mode
        * ``I``: Insert mode
        * ``v``: Visual mode

        Returns:
            A character representing the current mode
        """
        # TODO - sort this out
        if self.in_edit_mode():
            if insert_mode():
                return "I"
            elif replace_mode():
                return "o"
            else:
                return ">"
        return "^"

    def select(
        self,
        index: int,
        extend: bool = False,
        position: int | None = None,
        scroll: bool = True,
    ) -> None:
        """Select a cell or adds it to the selection.

        Args:
            index: The index of the cell to select
            extend: If true, the selection will be extended to include the cell
            position: An optional cursor position index to apply to the cell input
            scroll: Whether to scroll the page

        """
        self.page.select(index=index, extend=extend, position=position, scroll=scroll)
        # Focus the selected cell - use the current slice in case it did not change
        self.rendered_cells()[self.page.selected_slice.start].focus(
            position, scroll=False
        )

    def scroll_to(self, index: int) -> None:
        """Scroll to a cell by index."""
        self.page.scroll_to(index)

    def refresh(self, slice_: slice | None = None, scroll: bool = True) -> None:
        """Refresh the rendered contents of this notebook."""
        if slice_ is None:
            slice_ = self.page._selected_slice
        # This triggers another redraw of the selected cell
        self.page._set_selected_slice(slice_, force=True, scroll=scroll)
        self.page.reset()

    def refresh_cell(self, cell: Cell) -> None:
        """Trigger the refresh of a notebook cell."""
        self.page.get_child(cell.index).invalidate()

    def add_cell_above(self) -> None:
        """Inert a cell above the current selection."""
        index = self.page.selected_slice.start + 0
        self.add(index)
        self.refresh(slice_=slice(index, index + 1), scroll=False)

    def add_cell_below(self) -> None:
        """Inert a cell below the current selection."""
        index = self.page.selected_slice.start + 1
        self.add(index)
        self.refresh(slice_=slice(index, index + 1), scroll=True)

    def set_next_input(self, text: str, replace: bool = False) -> None:
        """Handle ``set_next_input`` payloads, e.g. ``%load`` magic."""
        if replace:
            self.cell.input_box.text = text
        else:
            index = self.page.selected_slice.start + 1
            self.add(index, source=text)
            self.refresh(slice_=slice(index, index + 1), scroll=True)

    def cut(self, slice_: slice | None = None) -> None:
        """Remove cells from the notebook and them to the `Notebook`'s clipboard."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        self.copy(slice_)
        self.delete(slice_)

    def copy(self, slice_: slice | None = None) -> None:
        """Add a copy of the selected cells to the `Notebook`'s clipboard."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        indices = range(*slice_.indices(len(self.json["cells"])))
        self.clipboard = deepcopy(
            # Sort clipboard contents by index)
            [
                x
                for _, x in sorted(
                    zip(indices, self.json["cells"][slice_]), key=lambda x: x[0]
                )
            ]
        )

    def copy_outputs(self, slice_: slice | None = None) -> None:
        """Copy the outputs of the selected cells."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        output_strings = []

        indices = sorted(range(*slice_.indices(len(self.json["cells"]))))
        rendered_cells = list(self._rendered_cells.values())

        for index in indices:
            cell = rendered_cells[index]
            output_strings.append(cell.output_area.to_plain_text())

        if output_strings:
            self.app.clipboard.set_data(ClipboardData("\n\n".join(output_strings)))

    def paste(self, index: int | None = None) -> None:
        """Append the contents of the `Notebook`'s clipboard below the current cell."""
        if index is None:
            index = self.page.selected_slice.start
        cell_jsons: MutableSequence = deepcopy(self.clipboard)
        # Assign a new cell IDs
        for cell_json in cell_jsons:
            cell_json["id"] = nbformat.v4.new_code_cell().get("id")
        self.json["cells"][index + 1 : index + 1] = cell_jsons
        self.dirty = True
        # Only change the selected cell if we actually pasted something
        if cell_jsons:
            self.refresh(slice(index + 1, index + 1 + len(cell_jsons)))

    def add(self, index: int, source: str = "", **kwargs: Any) -> None:
        """Create a new cell at a given index.

        Args:
            index: The position at which to insert a new cell
            source: The contents of the new cell
            kwargs: Additional parameters for the cell

        """
        self.json["cells"].insert(
            index,
            nbformat.v4.new_code_cell(source=source, **kwargs),
        )
        self.dirty = True

    def move(self, n: int, slice_: slice | None = None) -> None:
        """Move a slice of cells up or down.

        Args:
            slice_: A slice describing the cell indices to move
            n: The amount to move them by

        """
        if slice_ is None:
            slice_ = self.page.selected_slice
        if slice_ is not None:
            indices = range(*slice_.indices(len(self.json["cells"])))
            index = min(indices) + n
            if index >= 0 and index + len(indices) <= len(self.json["cells"]):
                cells = [
                    x
                    for _, x in sorted(
                        zip(indices, self.json["cells"][slice_]), key=lambda x: x[0]
                    )
                ]
                del self.json["cells"][slice_]
                self.json["cells"][index:index] = cells
                self.refresh(
                    slice(
                        slice_.start + n,
                        (-1 if slice_.stop is None else slice_.stop) + n,
                        slice_.step,
                    )
                )
            self.dirty = True

    def delete(self, slice_: slice | None = None) -> None:
        """Delete cells from the notebook."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        if slice_ is not None:
            indices = sorted(range(*slice_.indices(len(self.json["cells"]))))
            index = min(indices)
            cell_jsons = [
                x[1] for x in sorted(zip(indices, self.json["cells"][slice_]))
            ]
            self.undo_buffer.append((index, cell_jsons))
            del self.json["cells"][slice_]
            # Ensure there is always one cell
            if len(self.json["cells"]) == 0:
                self.add(1)
            # Get top cell of deleted range or cell above it
            slice_top = max(0, min(index, len(self.json["cells"]) - 1))
            self.refresh(slice(slice_top, slice_top + 1), scroll=True)
            self.dirty = True

    def undelete(self) -> None:
        """Insert the last deleted cell(s) back into the notebook."""
        if self.undo_buffer:
            index, cells = self.undo_buffer.pop()
            self.json["cells"][index:index] = cells
            self.refresh(slice(index, index + len(cells)))

    def merge(self, slice_: slice | None = None) -> None:
        """Merge two or more cells."""
        if slice_ is None:
            slice_ = self.page.selected_slice
        if slice_ is not None:
            indices = sorted(range(*slice_.indices(len(self.json["cells"]))))
            if len(indices) >= 2:
                # Create a new cell
                new_cell_json = nbformat.v4.new_code_cell()
                # Set the type for that for the focused cell
                cell_type = self.json["cells"][slice_.start].get("cell_type", "code")
                new_cell_json["cell_type"] = cell_type
                # Create and set the combined cell source
                sources = []
                for i in indices:
                    cell_json = self.json["cells"][i]
                    source = cell_json.get("source", "")
                    # Comment markdown cell contents if merging into code cell
                    if cell_type == "code" and cell_json.get("cell_type") == "markdown":
                        source = "\n".join([f"# {line}" for line in source.split("\n")])
                    sources.append(source)
                new_cell_json["source"] = "\n\n".join(sources)
                # Insert the new cell
                new_index = max(indices) + 1
                self.json["cells"].insert(new_index, new_cell_json)
                # Delete the selected slice
                self.delete(slice_)
                # Select the new cell
                self.select(indices[0])
                self.dirty = True

    def split_cell(self) -> None:
        """Split a cell into two at the cursor position."""
        # Get the current cell
        cell = self.cell
        # Get the cursor position in the cell
        cursor_position = cell.input_box.buffer.cursor_position
        # Reset the cursor position to avoid an error after splitting
        cell.input_box.buffer.cursor_position = 0
        # Get the cell contents
        source = cell.json.get("source", "")
        # Create a new cell
        new_cell_json = nbformat.v4.new_code_cell()
        # Split the cell contents at the cursor position
        cell.input = source[cursor_position:]
        new_cell_json["source"] = source[:cursor_position]
        # Copy the cell type
        new_cell_json["cell_type"] = cell.json["cell_type"]
        # Add the new cell to the notebook
        self.json["cells"].insert(cell.index, new_cell_json)
        # Refresh the notebook display
        self.refresh()

    def reformat(self) -> None:
        """Reformat all code cells in the notebooks."""
        for cell in self.rendered_cells():
            if cell.cell_type == "code":
                cell.input_box.reformat()

    def run_selected_cells(
        self,
        advance: bool = False,
        insert: bool = False,
    ) -> None:
        """Run the currently selected cells.

        Args:
            advance: If True, move to next cell. If True and at the last cell, create a
                new cell at the end of the notebook.
            insert: If True, add a new empty cell below the current cell and select it.

        """
        self.exit_edit_mode()
        n_cells = len(self.json["cells"])
        selected_indices = self.page.selected_indices
        cells = {cell.index: cell for cell in self._rendered_cells.values()}

        # Run the cells
        for i in sorted(selected_indices):
            cells[i].run_or_render()
        # Insert a cell if we are at the last cell
        index = max(selected_indices)
        if insert or (advance and max(selected_indices) == (n_cells) - 1):
            self.add(index + 1)
            self.refresh(slice_=slice(index + 1, index + 2), scroll=True)
        elif advance:
            self.select(index + 1)

    def run_all(self, wait: bool = False) -> None:
        """Run all cells."""
        if self.kernel:
            log.debug("Running all cells (wait=%s)", wait)
            for cell in self.rendered_cells():
                if cell.json.get("cell_type") == "code":
                    log.debug("Running cell %s", cell.id)
                    cell.run_or_render(wait=wait)
            log.debug("All cells run")

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.notebook.tabs.notebook.Notebook": {
                "enter-cell-edit-mode": "enter",
                "exit-edit-mode": "escape",
                "run-selected-cells": ["c-enter", "c-e", "c-s-f8"],
                "run-and-select-next": ["s-enter", "c-r", "s-f9"],
                "run-cell-and-insert-below": ("A-enter"),
                "add-cell-above": "a",
                "add-cell-below": "b",
                "delete-cells": ("d", "d"),
                "undelete-cells": "z",
                "cut-cells": "x",
                "copy-cells": "c",
                "copy-outputs": ("A-c"),
                "paste-cells": "v",
                "interrupt-kernel": ("i", "i"),
                "restart-kernel": ("0", "0"),
                "scroll-up": ["[", "<scroll-up>"],
                "scroll-down": ["]", "<scroll-down>"],
                "scroll-up-5-lines": "{",
                "scroll-down-5-lines": "}",
                "select-first-cell": ["home", "c-up"],
                "select-5th-previous-cell": "pageup",
                "select-previous-cell": ["up", "k"],
                "select-next-cell": ["down", "j"],
                "select-5th-next-cell": "pagedown",
                "select-last-cell": ["end", "c-down"],
                "select-all-cells": "c-a",
                "extend-cell-selection-to-top": "s-home",
                "extend-cell-selection-up": ["s-up", "K"],
                "extend-cell-selection-down": ["s-down", "J"],
                "extend-cell-selection-to-bottom": "s-end",
                "move-cells-up": ("A-up"),
                "move-cells-down": ("A-down"),
                "cells-to-markdown": "m",
                "cells-to-code": "y",
                "cells-to-raw": "r",
                "reformat-cells": "f",
                "reformat-notebook": "F",
                "edit-in-external-editor": "e",
                "merge-cells": "M",
                "split-cell": "c-\\",
                "edit-previous-cell": "up",
                "edit-next-cell": "down",
                "edit-previous-cell-vi": "k",
                "edit-next-cell-vi": "j",
                "scroll-output-left": "left",
                "scroll-output-right": "right",
                "toggle-expand": "w",
                "toggle-wrap-cell-outputs": "W",
                "notebook-toggle-line-numbers": "l",
            }
        }
    )



"""Overrides for PTK containers which only render visible lines."""

from __future__ import annotations

import logging

from prompt_toolkit.layout import screen

from euporie.core.data_structures import DiInt

log = logging.getLogger(__name__)


class BoundedWritePosition(screen.WritePosition):
    """A write position which also hold bounding box information."""

    def __init__(
        self,
        xpos: int,
        ypos: int,
        width: int,
        height: int,
        bbox: DiInt | None = None,
    ) -> None:
        """Create a new instance of the write position."""
        super().__init__(xpos, ypos, width, height)
        self.bbox = bbox or DiInt(0, 0, 0, 0)

    def __repr__(self) -> str:
        """Return a string representation of the write position."""
        return (
            f"{self.__class__.__name__}("
            f"x={self.xpos}, y={self.ypos}, "
            f"w={self.width}, h={self.height}, "
            f"bbox={self.bbox})"
        )


class Screen(screen.Screen):
    """Screen class which uses :py:`BoundedWritePosition`s."""

    def fill_area(
        self, write_position: screen.WritePosition, style: str = "", after: bool = False
    ) -> None:
        """Fill the content of this area, using the given `style`."""
        if not style.strip():
            return

        if isinstance(write_position, BoundedWritePosition):
            bbox = write_position.bbox
        else:
            bbox = DiInt(0, 0, 0, 0)

        xmin = write_position.xpos + bbox.left
        xmax = write_position.xpos + write_position.width - bbox.right
        char_cache = screen._CHAR_CACHE
        data_buffer = self.data_buffer

        if after:
            append_style = " " + style
            prepend_style = ""
        else:
            append_style = ""
            prepend_style = style + " "

        for y in range(
            write_position.ypos + bbox.top,
            write_position.ypos + write_position.height - bbox.bottom,
        ):
            row = data_buffer[y]
            for x in range(xmin, xmax):
                cell = row[x]
                row[x] = char_cache[
                    cell.char, prepend_style + cell.style + append_style
                ]



"""Define commands for widgets."""

from euporie.core.commands import add_cmd


@add_cmd()
def _toggle_side_bar_pane() -> None:
    """Open or close the current side-bar pane."""
    from euporie.notebook.current import get_app

    get_app().side_bar.toggle_pane()



"""Allow access to the current running application."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.application.current import _current_app_session

if TYPE_CHECKING:
    from euporie.core.app.app import BaseApp


def get_app() -> BaseApp:
    """Get the current active (running) Application."""
    from euporie.core.app.app import BaseApp

    session = _current_app_session.get()
    if isinstance(session.app, BaseApp):
        return session.app

    # Create a dummy application if we really need one
    from euporie.core.app.dummy import DummyApp

    return DummyApp()


def get_app_cls(name: str) -> BaseApp:
    """Load a euporie app by name."""
    from euporie.core.__main__ import available_apps

    return available_apps()[name].load()



"""Contain a tab for displaying JSON data."""

from __future__ import annotations

import asyncio
import json
import logging
from typing import TYPE_CHECKING, ClassVar

from prompt_toolkit.layout.containers import VSplit
from prompt_toolkit.layout.dimension import Dimension

from euporie.core.layout.scroll import ScrollingContainer
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.tabs.base import Tab
from euporie.core.widgets.tree import JsonView

if TYPE_CHECKING:
    from pathlib import Path

    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.app.app import BaseApp
    from euporie.core.bars.status import StatusBarFields

log = logging.getLogger(__name__)


class JsonTab(Tab):
    """Tab class for JSON data."""

    name = "JSON Viewer"
    mime_types: ClassVar[set[str]] = {"*json"}
    file_types: ClassVar[dict[str, None]] = {".json": None}

    def __init__(self, app: BaseApp, path: Path | None = None) -> None:
        """Call when the tab is created."""
        super().__init__(app, path)

        # Load file and container in background
        if self.path is not None:

            def _load() -> None:
                self.container = self.load_container()
                self.app.layout.focus(self.container)
                self.app.invalidate()

            app.create_background_task(asyncio.to_thread(_load))

    def __pt_status__(self) -> StatusBarFields | None:
        """Return a list of statusbar field values shown then this tab is active."""
        return ([str(self.path)], [])

    @property
    def title(self) -> str:
        """Return the tab title."""
        if self.path is not None:
            return str(self.path.name) or str(self.path)
        else:
            return "<file>"

    def load_container(self) -> AnyContainer:
        """Abcract method for loading the notebook's main container."""
        assert self.path is not None

        data = json.load(self.path.open())

        return VSplit(
            [
                scroll := ScrollingContainer(
                    children=[JsonView(data, title=self.path.name, expanded=True)]
                ),
                MarginContainer(ScrollbarMargin(), target=scroll),
            ],
            width=Dimension(weight=1),
            height=Dimension(weight=1),
        )



"""Contain kernel tab base class."""

from __future__ import annotations

import asyncio
import logging
from abc import ABCMeta
from collections import deque
from functools import lru_cache, partial
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from prompt_toolkit.auto_suggest import DummyAutoSuggest, DynamicAutoSuggest
from prompt_toolkit.completion.base import (
    DynamicCompleter,
    _MergedCompleter,
)
from prompt_toolkit.history import DummyHistory, InMemoryHistory

from euporie.core.app.current import get_app
from euporie.core.comm.registry import open_comm
from euporie.core.commands import add_cmd
from euporie.core.completion import DeduplicateCompleter, KernelCompleter, LspCompleter
from euporie.core.diagnostics import Report
from euporie.core.filters import kernel_tab_has_focus
from euporie.core.format import LspFormatter
from euporie.core.history import KernelHistory
from euporie.core.inspection import (
    FirstInspector,
    KernelInspector,
    LspInspector,
)
from euporie.core.kernel import list_kernels
from euporie.core.kernel.base import NoKernel
from euporie.core.tabs.base import Tab

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path
    from typing import Any, Callable

    from prompt_toolkit.auto_suggest import AutoSuggest
    from prompt_toolkit.completion.base import Completer
    from prompt_toolkit.history import History

    from euporie.core.app.app import BaseApp
    from euporie.core.comm.base import Comm
    from euporie.core.format import Formatter
    from euporie.core.inspection import Inspector
    from euporie.core.kernel.base import BaseKernel, KernelFactory, MsgCallbacks
    from euporie.core.lsp import LspClient
    from euporie.core.widgets.inputs import KernelInput

log = logging.getLogger(__name__)


@lru_cache
def autosuggest_factory(kind: str, history: History) -> AutoSuggest:
    """Generate autosuggesters."""
    if kind == "smart":
        from euporie.core.suggest import SmartHistoryAutoSuggest

        return SmartHistoryAutoSuggest(history)
    elif kind == "simple":
        from euporie.core.suggest import SimpleHistoryAutoSuggest

        return SimpleHistoryAutoSuggest(history)
    else:
        from prompt_toolkit.auto_suggest import DummyAutoSuggest

        return DummyAutoSuggest()


class KernelTab(Tab, metaclass=ABCMeta):
    """A Tab which connects to a kernel."""

    kernel_language: str
    _metadata: dict[str, Any]
    bg_init = False

    default_callbacks: MsgCallbacks
    allow_stdin: bool

    def __init__(
        self,
        app: BaseApp,
        path: Path | None = None,
        kernel: BaseKernel | None = None,
        comms: dict[str, Comm] | None = None,
        use_kernel_history: bool = False,
        connection_file: Path | None = None,
    ) -> None:
        """Create a new instance of a tab with a kernel."""
        # Init tab
        super().__init__(app, path)

        self.kernel: BaseKernel = NoKernel(self)
        self.lsps: list[LspClient] = []
        self.history: History = DummyHistory()
        self.inspectors: list[Inspector] = []
        self.inspector = FirstInspector(lambda: self.inspectors)
        self.suggester: AutoSuggest = DummyAutoSuggest()
        self.completers: list[Completer] = []
        self.completer = DeduplicateCompleter(
            DynamicCompleter(lambda: _MergedCompleter(self.completers))
        )
        self.formatters: list[Formatter] = self.app.formatters
        self.reports: WeakKeyDictionary[LspClient, Report] = WeakKeyDictionary()

        # The client-side comm states
        self.comms: dict[str, Comm] = {}
        # The current kernel input
        self._current_input: KernelInput | None = None

        if self.bg_init:
            # Load kernel in a background thread
            app.create_background_task(
                asyncio.to_thread(
                    self.init_kernel, kernel, comms, use_kernel_history, connection_file
                )
            )
        else:
            self.init_kernel(kernel, comms, use_kernel_history, connection_file)

    async def load_lsps(self) -> None:
        """Load the LSP clients."""
        path = self.path

        # Load list of LSP clients for the tab's language
        self.lsps.extend(self.app.get_language_lsps(self.language))

        # Wait for all lsps to be initialized, and setup hooks as they become ready
        async def _await_load(lsp: LspClient) -> LspClient:
            await lsp.initialized.wait()
            return lsp

        for ready in asyncio.as_completed([_await_load(lsp) for lsp in self.lsps]):
            lsp = await ready
            # Apply open, save, and close hooks to the tab
            change_handler = partial(lambda lsp, tab: self.lsp_change_handler(lsp), lsp)
            close_handler = partial(lambda lsp, tab: self.lsp_close_handler(lsp), lsp)
            before_save_handler = partial(
                lambda lsp, tab: self.lsp_before_save_handler(lsp), lsp
            )
            after_save_handler = partial(
                lambda lsp, tab: self.lsp_after_save_handler(lsp), lsp
            )

            self.on_close += close_handler
            self.on_change += change_handler
            self.before_save += before_save_handler
            self.after_save += after_save_handler

            # Listen for LSP diagnostics
            lsp.on_diagnostics += self.lsp_update_diagnostics

            # Add completer
            completer = LspCompleter(lsp=lsp, path=path)
            self.completers.append(completer)

            # Add inspector
            inspector = LspInspector(lsp, path)
            self.inspectors.append(inspector)

            # Add formatter
            formatter = LspFormatter(lsp, path)
            self.formatters.append(formatter)

            # Remove hooks if the LSP exits
            def lsp_unload(lsp: LspClient) -> None:
                self.on_change -= change_handler  # noqa: B023
                self.before_save -= before_save_handler  # noqa: B023
                self.after_save -= after_save_handler  # noqa: B023
                self.on_close -= close_handler  # noqa: B023
                if completer in self.completers:  # noqa: B023
                    self.completers.remove(completer)  # noqa: B023
                if inspector in self.completers:  # noqa: B023
                    self.inspectors.remove(inspector)  # noqa: B023
                if formatter in self.completers:  # noqa: B023
                    self.formatters.remove(formatter)  # noqa: B023
                if completer in self.completers:  # noqa: B023
                    self.completers.remove(completer)  # noqa: B023
                if inspector in self.inspectors:  # noqa: B023
                    self.inspectors.remove(inspector)  # noqa: B023
                if formatter in self.formatters:  # noqa: B023
                    self.formatters.remove(formatter)  # noqa: B023

            lsp.on_exit += lsp_unload

            # Remove the lsp exit handler if this tab closes
            self.on_close += lambda tab: (
                (lsp.on_exit.__isub__(lsp_unload) and None) or None  # noqa: B023
            )  # Magical typing

            # Tell the LSP we have an open file
            self.lsp_open_handler(lsp)

    def pre_init_kernel(self) -> None:
        """Run stuff before the kernel is loaded."""

    def post_init_kernel(self) -> None:
        """Run stuff after the kernel is loaded."""
        if not isinstance(self.kernel, NoKernel):
            self.metadata["kernelspec"] = self.kernel.spec

    def init_kernel(
        self,
        kernel: BaseKernel | None = None,
        comms: dict[str, Comm] | None = None,
        use_kernel_history: bool = False,
        connection_file: Path | None = None,
    ) -> None:
        """Set up the tab's kernel and related components."""
        self.pre_init_kernel()

        self.kernel_queue: deque[Callable] = deque()

        if kernel:
            self.kernel = kernel
            self.kernel.default_callbacks = self.default_callbacks
        else:
            from euporie.core.kernel import list_kernels

            kernel_infos = list_kernels()
            kernel_name = self.kernel_name or self.app.config.kernel_name
            for info in kernel_infos:
                if info.name == kernel_name:
                    factory = info.factory
                    break
            else:
                msg = (
                    f"Kernel '{self.kernel_display_name}' not found"
                    if self.kernel_name
                    else "No kernel selected"
                )
                self.change_kernel(msg=msg, startup=True)
                return
            self.kernel = factory(
                kernel_tab=self,
                allow_stdin=self.allow_stdin,
                default_callbacks=self.default_callbacks,
                **(
                    {"connection_file": connection_file}
                    if connection_file is not None
                    else {}
                ),
            )

        self.comms = comms or {}  # The client-side comm states
        self.completers.append(KernelCompleter(lambda: self.kernel))
        self.inspectors.append(KernelInspector(lambda: self.kernel))
        self.use_kernel_history = use_kernel_history
        self.history = (
            KernelHistory(lambda: self.kernel)
            if use_kernel_history
            else InMemoryHistory()
        )

        def _get_suggester() -> AutoSuggest | None:
            return autosuggest_factory(self.app.config.autosuggest, self.history)

        self.suggester = DynamicAutoSuggest(_get_suggester)

        self.app.create_background_task(self.load_lsps())

        self.post_init_kernel()

    def close(self, cb: Callable | None = None) -> None:
        """Shut down kernel when tab is closed."""
        if self.kernel is not None:
            self.kernel.shutdown()
        super().close(cb)

    def interrupt_kernel(self) -> None:
        """Interrupt the current `Notebook`'s kernel."""
        self.kernel.interrupt()

    def restart_kernel(self, cb: Callable | None = None) -> None:
        """Restart the current `Notebook`'s kernel."""

        def _cb(result: dict[str, Any]) -> None:
            self.kernel_started()
            if callable(cb):
                cb()

        if confirm := self.app.dialogs.get("confirm"):
            confirm.show(
                message="Are you sure you want to restart the kernel?",
                cb=partial(self.kernel.restart, cb=_cb),
            )
        else:
            self.kernel.restart(cb=_cb)

    def kernel_started(self, result: dict[str, Any] | None = None) -> None:
        """Task to run when the kernel has started."""
        # Set kernel spec in metadata
        if self.kernel.status == "error":
            self.report_kernel_error(self.kernel.error)

        else:
            # Wait for an idle kernel
            if self.kernel.status != "idle":
                self.kernel.wait_for_status("idle")

            # Load widget comm info
            # self.kernel.comm_info(target_name="jupyter.widget")

            # Load kernel info
            self.kernel.info(set_kernel_info=self.set_kernel_info)

            # Load kernel history
            if self.use_kernel_history:
                self.app.create_background_task(self.load_history())

            # Run queued kernel tasks when the kernel is idle
            log.debug("Running %d kernel tasks", len(self.kernel_queue))
            while self.kernel_queue:
                self.kernel_queue.popleft()()

        self.app.invalidate()

    def report_kernel_error(self, error: Exception | None) -> None:
        """Report a kernel error to the user."""
        log.debug("Kernel error", exc_info=error)

    async def load_history(self) -> None:
        """Load kernel history."""
        try:
            await self.history.load().__anext__()
        except StopAsyncIteration:
            pass

    @property
    def metadata(self) -> dict[str, Any]:
        """Return a dictionary to hold notebook / kernel metadata."""
        return self._metadata

    @property
    def kernel_name(self) -> str:
        """Return the name of the kernel defined in the notebook JSON."""
        return self.metadata.get("kernelspec", {}).get("name", "")

    @kernel_name.setter
    def kernel_name(self, value: str) -> None:
        """Return the name of the kernel defined in the notebook JSON."""
        self.metadata.setdefault("kernelspec", {})["name"] = value

    @property
    def language(self) -> str:
        """Return the name of the kernel defined in the notebook JSON."""
        return self.metadata.get("kernelspec", {}).get("language")

    @property
    def kernel_display_name(self) -> str:
        """Return the display name of the kernel defined in the notebook JSON."""
        return self.metadata.get("kernelspec", {}).get("display_name", self.kernel_name)

    @property
    def kernel_lang_file_ext(self) -> str:
        """Return the display name of the kernel defined in the notebook JSON."""
        return self.metadata.get("language_info", {}).get("file_extension", ".py")

    @property
    def current_input(self) -> KernelInput:
        """Return the currently active kernel input, if any."""
        from euporie.core.widgets.inputs import KernelInput

        return self._current_input or KernelInput(self)

    def set_kernel_info(self, info: dict) -> None:
        """Handle kernel info requests."""
        self.metadata["language_info"] = info.get("language_info", {})

    def change_kernel(self, msg: str | None = None, startup: bool = False) -> None:
        """Prompt the user to select a new kernel."""
        kernel_infos = list_kernels()

        # Warn the user if no kernels are installed
        if not kernel_infos:
            if startup and "no-kernels" in self.app.dialogs:
                self.app.dialogs["no-kernels"].show()
            return

        # Automatically select the only kernel if there is only one
        if startup and len(kernel_infos) == 1:
            self.switch_kernel(next(iter(kernel_infos)).factory)
            return

        # Prompt user to select a kernel
        self.app.dialogs["change-kernel"].show(tab=self, message=msg)

    def switch_kernel(self, factory: KernelFactory) -> None:
        """Shut down the current kernel and change to another."""
        if (old_kernel := self.kernel) is not None:
            old_kernel.shutdown(wait=True)
        kernel = factory(
            kernel_tab=self,
            default_callbacks=self.default_callbacks,
            allow_stdin=self.allow_stdin,
        )
        self.init_kernel(kernel)

    def comm_open(self, content: dict, buffers: Sequence[bytes]) -> None:
        """Register a new kernel Comm object in the notebook."""
        comm_id = str(content.get("comm_id"))
        self.comms[comm_id] = open_comm(
            comm_container=self, content=content, buffers=buffers
        )

    def comm_msg(self, content: dict, buffers: Sequence[bytes]) -> None:
        """Respond to a Comm message from the kernel."""
        comm_id = str(content.get("comm_id"))
        if comm := self.comms.get(comm_id):
            comm.process_data(content.get("data", {}), buffers)

    def comm_close(self, content: dict, buffers: Sequence[bytes]) -> None:
        """Close a notebook Comm."""
        comm_id = content.get("comm_id")
        if comm_id in self.comms:
            del self.comms[comm_id]

    def lsp_open_handler(self, lsp: LspClient) -> None:
        """Tell the LSP we opened a file."""
        lsp.open_doc(
            path=self.path, language=self.language, text=self.current_input.buffer.text
        )

    def lsp_change_handler(self, lsp: LspClient) -> None:
        """Tell the LSP server a file has changed."""
        lsp.change_doc(
            path=self.path,
            language=self.language,
            text=self.current_input.buffer.text,
        )

    def lsp_before_save_handler(self, lsp: LspClient) -> None:
        """Tell the the LSP we are about to save a document."""
        lsp.will_save_doc(self.path)

    def lsp_after_save_handler(self, lsp: LspClient) -> None:
        """Tell the the LSP we saved a document."""
        lsp.save_doc(self.path, text=self.current_input.buffer.text)

    def lsp_close_handler(self, lsp: LspClient) -> None:
        """Tell the LSP we opened a file."""
        lsp.close_doc(path=self.path)

    def lsp_update_diagnostics(self, lsp: LspClient) -> None:
        """Process a new diagnostic report from the LSP."""
        if (diagnostics := lsp.reports.pop(self.path.as_uri(), None)) is not None:
            self.reports[lsp] = Report.from_lsp(self.current_input.text, diagnostics)
            self.app.invalidate()

    def report(self) -> Report:
        """Return the current diagnostic reports."""
        return Report.from_reports(*self.reports.values())

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=kernel_tab_has_focus)
    def _change_kernel() -> None:
        """Change the notebook's kernel."""
        if isinstance(kt := get_app().tab, KernelTab):
            kt.change_kernel()



"""Contains container classes for diagnostic information."""

from __future__ import annotations

import logging
from abc import ABCMeta
from typing import TYPE_CHECKING, NamedTuple

if TYPE_CHECKING:
    from typing import Literal

log = logging.getLogger(__name__)


class Diagnostic(NamedTuple):
    """Represent a diagnostic item for code."""

    code: str
    message: str
    level: Literal[0, 1, 2, 3, 4, 5]  # Like Python logging debug levels
    link: str
    lines: slice
    chars: slice


class Report(list[Diagnostic], metaclass=ABCMeta):
    """Class for storing a diagnostic report."""

    @classmethod
    def from_lsp(cls, text: str, outputs: list[dict]) -> Report:
        """Create a diagnostic report from LSP output."""
        items = []
        for output in outputs:
            # Calculate slice from range
            range_ = output.get("range", {})
            start = range_.get("start", {})
            end = range_.get("end", {})
            desc = output.get("codeDescription", {})
            items.append(
                Diagnostic(
                    code=output.get("code", ""),
                    message=output.get("message", ""),
                    level=5 - output.get("severity", 2),
                    link=desc.get("href"),
                    lines=slice(start.get("line"), end.get("line") + 1),
                    chars=slice(start.get("character"), end.get("character") + 1),
                )
            )
        return cls(items)

    @classmethod
    def from_reports(cls, *reports: Report) -> Report:
        """Initialize a new report."""
        return cls([diagnostic for report in reports for diagnostic in report])



"""Define settings for the notebook app."""

from euporie.core.config import add_setting
from euporie.notebook.enums import TabMode

add_setting(
    name="tab_mode",
    group="euporie.notebook.app",
    flags=["--tab-mode"],
    type_=str,
    choices=[mode.value for mode in TabMode],
    default="stack",
    help_="The method used to display multiple tabs",
    description="""
        Determines how multiple tabs are displayed when more than one tab is open.
        * ``stack`` displays one tab at a time with a tab-bar
        * ``tile_horizontally`` displays tabs side-by-side
        * ``tile_vertically`` displays tabs one-atop-the-next
    """,
)

add_setting(
    name="always_show_tab_bar",
    group="euporie.notebook.app",
    flags=["--always-show-tab-bar"],
    type_=bool,
    help_="Always show the tab bar",
    default=False,
    description="""
        When set, the tab bar will always be shown - otherwise the tab bar is only
        shown when multiple tabs are open.
    """,
)

add_setting(
    name="background_pattern",
    group="euporie.notebook.app",
    flags=["--background-pattern", "--bg-pattern"],
    type_=int,
    choices=list(range(6)),
    help_="The background pattern to use",
    default=2,
    schema={
        "minimum": 0,
        "maximum": 5,
    },
    description="""
        The background pattern to use when the notebook is narrower than the
        available width. Zero mean no pattern is used.
    """,
)

add_setting(
    name="background_character",
    group="euporie.notebook.app",
    flags=["--background-character", "--bg-char"],
    type_=str,
    help_="Character for background pattern",
    default="·",
    schema={
        "maxLength": 1,
    },
    description="""
        The character to use when drawing the background pattern.

        Recommended characters include: "·", "⬤", "╳", "╱", "╲", "░", "▒", "▓", "▞", "╬"
    """,  # ,
)

add_setting(
    name="run_after_external_edit",
    group="euporie.notebook.app",
    flags=["--run-after-external-edit"],
    type_=bool,
    help_="Run cells after editing externally",
    default=False,
    description="""
        Whether to execute a cell immediately after editing in `$EDITOR`.
    """,
)

add_setting(
    name="run",
    group="euporie.notebook.app",
    flags=["--run"],
    type_=bool,
    help_="Run the notebook files when loaded",
    default=False,
    description="""
        If set, notebooks will be run automatically when opened, or if previewing a
        file, the notebooks will be run before being output.
    """,
)

add_setting(
    name="show_top_bar",
    group="euporie.notebook.app",
    flags=["--show-top-bar"],
    type_=bool,
    title="top bar",
    help_="Show the top bar",
    default=True,
    schema={"type": "boolean"},
    description="""
        Whether the top bar should be shown at the top of the screen.
    """,
)



"""Miscellaneous utility classes."""

from __future__ import annotations

from collections.abc import Sequence
from functools import cache
from itertools import chain
from typing import TYPE_CHECKING, TypeVar, overload

from prompt_toolkit.mouse_events import MouseButton, MouseEventType

if TYPE_CHECKING:
    from collections.abc import Iterable
    from types import ModuleType
    from typing import Callable

    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.mouse_handlers import MouseHandler
    from prompt_toolkit.mouse_events import MouseEvent

T = TypeVar("T")


class ChainedList(Sequence[T]):
    """A list-like class which chains multiple lists."""

    def __init__(self, *lists: Iterable[T]) -> None:
        """Create a new instance."""
        self.lists = lists

    @property
    def data(self) -> list[T]:
        """Return the list data."""
        return list(chain.from_iterable(self.lists))

    @overload
    def __getitem__(self, i: int) -> T: ...

    @overload
    def __getitem__(self, i: slice) -> list[T]: ...

    def __getitem__(self, i):
        """Get an item from the chained lists."""
        return self.data[i]

    def __len__(self) -> int:
        """Return the length of the chained lists."""
        return len(self.data)


def dict_merge(target_dict: dict, input_dict: dict) -> None:
    """Merge the second dictionary onto the first."""
    for k in input_dict:
        if k in target_dict:
            if isinstance(target_dict[k], dict) and isinstance(input_dict[k], dict):
                dict_merge(target_dict[k], input_dict[k])
            elif isinstance(target_dict[k], list) and isinstance(input_dict[k], list):
                target_dict[k] = [*target_dict[k], *input_dict[k]]
            else:
                target_dict[k] = input_dict[k]
        else:
            target_dict[k] = input_dict[k]


def on_click(func: Callable) -> MouseHandler:
    """Return a mouse handler which call a given function on click."""

    def _mouse_handler(mouse_event: MouseEvent) -> NotImplementedOrNone:
        if (
            mouse_event.button == MouseButton.LEFT
            and mouse_event.event_type == MouseEventType.MOUSE_UP
        ):
            return func()
        return NotImplemented

    return _mouse_handler


@cache
def root_module(name: str) -> ModuleType:
    """Find and load the root module of a given module name by traversing up the module hierarchy.

    This function walks up the module hierarchy until it finds the topmost parent module
    that has a valid location. It uses Python's importlib machinery to inspect module
    specifications and load modules.

    Args:
        name: The name of the module to find the root for (e.g., 'package.subpackage.module')

    Returns:
        The loaded root module object

    Example:
        >>> root = root_module("django.contrib.admin")
        >>> print(root.__name__)
        'django'

    Note:
        The function is cached using lru_cache to improve performance for repeated lookups.
        The function handles both regular packages and frozen modules.
    """
    from importlib.util import find_spec, module_from_spec

    if spec := find_spec(name):
        while True:
            if spec.name == spec.parent:
                try:
                    parent = find_spec("..", spec.parent)
                except ImportError:
                    break
            elif spec.parent is not None:
                parent = find_spec(spec.parent)
            if (spec and not spec.parent) or (parent and not parent.has_location):
                break
            if parent:
                spec = parent
        if spec.loader:
            module = module_from_spec(spec)
            spec.loader.exec_module(module)
            return module
    raise ModuleNotFoundError(name=name)


@cache
def import_submodules(root: ModuleType, names: tuple[str]) -> list[ModuleType]:
    """Import all submodules with a specific name within a root module's package hierarchy.

    This function walks through all packages under the given root module and imports
    any submodules that match the specified name. It handles various module types
    including regular packages, single file modules, and frozen modules.

    Args:
        root: The root module object to search within
        names: The specific submodule name to search for

    Returns:
        A list of imported module objects matching the specified name

    Example:
        >>> root = import_module("django")
        >>> admin_modules = import_submodules(root, "admin")
        >>> print([m.__name__ for m in admin_modules])
        ['django.contrib.admin', 'django.contrib.gis.admin']

    Note:
        - The function is cached using lru_cache to improve performance for repeated imports
        - For packages, it searches through __path__
        - For single file modules, it uses __file__
        - For frozen modules, it uses the module specification's origin
    """
    from importlib import import_module
    from pkgutil import walk_packages

    if hasattr(root, "__path__"):
        path = root.__path__
    elif hasattr(root, "__file__"):
        path = [root.__file__] if root.__file__ else []
    else:
        # For frozen modules, we need to create a special path
        spec = root.__spec__
        path = [spec.origin] if spec and spec.origin else []
    return [
        import_module(module_name)
        for _loader, module_name, _is_pkg in walk_packages(path, f"{root.__name__}.")
        if module_name.rpartition(".")[2] in names
    ]



"""Define controls for display of terminal graphics."""

from __future__ import annotations

import logging
import weakref
from abc import ABCMeta, abstractmethod
from math import ceil, floor
from typing import TYPE_CHECKING

from prompt_toolkit.cache import FastDictCache, SimpleCache
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.formatted_text.utils import split_lines
from prompt_toolkit.layout.containers import Float, Window
from prompt_toolkit.layout.controls import GetLinePrefixCallable, UIContent, UIControl
from prompt_toolkit.layout.mouse_handlers import MouseHandlers
from prompt_toolkit.layout.screen import Char, WritePosition
from prompt_toolkit.utils import get_cwidth

from euporie.core.app.current import get_app
from euporie.core.convert.datum import Datum
from euporie.core.convert.registry import find_route
from euporie.core.data_structures import DiInt
from euporie.core.filters import has_float, in_mplex
from euporie.core.ft.utils import _ZERO_WIDTH_FRAGMENTS
from euporie.core.io import passthrough
from euporie.core.layout.scroll import BoundedWritePosition

if TYPE_CHECKING:
    from typing import Any, Callable, ClassVar

    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.layout.screen import Screen


log = logging.getLogger(__name__)


class GraphicControl(UIControl, metaclass=ABCMeta):
    """A base-class for display controls which render terminal graphics."""

    def __init__(
        self,
        datum: Datum,
        scale: float = 0,
        bbox: DiInt | None = None,
    ) -> None:
        """Initialize the graphic control."""
        self.app = get_app()
        self.datum = datum
        self.scale = scale
        self.bbox = bbox or DiInt(0, 0, 0, 0)
        self.rendered_lines: list[StyleAndTextTuples] = []
        self._content_cache: SimpleCache = SimpleCache(maxsize=50)
        self._format_cache: SimpleCache = SimpleCache(maxsize=50)

    def preferred_width(self, max_available_width: int) -> int | None:
        """Return the width of the rendered content."""
        cols, _aspect = self.datum.cell_size()
        return min(cols, max_available_width) if cols else max_available_width

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int:
        """Return the number of lines in the rendered content."""
        cols, aspect = self.datum.cell_size()
        if aspect:
            return ceil(min(width, cols) * aspect)
        self.rendered_lines = self.get_rendered_lines(width, max_available_height)
        return len(self.rendered_lines)

    @abstractmethod
    def convert_data(self, wp: WritePosition) -> str:
        """Convert datum to required format."""
        return ""

    @abstractmethod
    def get_rendered_lines(
        self, visible_width: int, visible_height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Render the output data."""
        return []

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate rendered output at a given size.

        Args:
            width: The desired output width
            height: The desired output height

        Returns:
            `UIContent` for the given output size.

        """

        def get_content() -> dict[str, Any]:
            rendered_lines = self.get_rendered_lines(width, height)
            self.rendered_lines = rendered_lines[:]
            line_count = len(rendered_lines)

            def get_line(i: int) -> StyleAndTextTuples:
                # Return blank lines if the renderer expects more content than we have
                line = []
                if i < line_count:
                    line += rendered_lines[i]
                # Add a space at the end, because that is a possible cursor position.
                # This is what PTK does, and fixes a nasty bug which took me ages to
                # track down the source of, where scrolling would stop working when the
                # cursor was on an empty line.
                line += [("", " ")]
                return line

            return {
                "get_line": get_line,
                "line_count": line_count,
                "menu_position": Point(0, 0),
            }

        # Re-render if the image width changes, or the terminal character size changes
        key = (
            width,
            height,
            self.app.color_palette,
            self.app.cell_size_px,
            self.bbox,
        )
        return UIContent(
            **self._content_cache.get(key, get_content),
        )

    def hide(self) -> None:
        """Hide the graphic from show."""

    def close(self) -> None:
        """Remove the displayed object entirely."""
        if not self.app.leave_graphics():
            self.hide()


class SixelGraphicControl(GraphicControl):
    """A graphic control which displays images as sixels."""

    def convert_data(self, wp: WritePosition) -> str:
        """Convert datum to required format."""
        bbox = wp.bbox if isinstance(wp, BoundedWritePosition) else DiInt(0, 0, 0, 0)
        cmd = str(self.datum.convert(to="sixel", cols=wp.width, rows=wp.height)).strip()
        if any(bbox):
            from sixelcrop import sixelcrop

            cell_size_x, cell_size_y = self.app.cell_size_px

            cmd = sixelcrop(
                data=cmd,
                # Horizontal pixel offset of the displayed image region
                x=bbox.left * cell_size_x,
                # Vertical pixel offset of the displayed image region
                y=bbox.top * cell_size_y,
                # Pixel width of the displayed image region
                w=(wp.width - bbox.left - bbox.right) * cell_size_x,
                # Pixel height of the displayed image region
                h=(wp.height - bbox.top - bbox.bottom) * cell_size_y,
            )

        return cmd

    def get_rendered_lines(
        self, visible_width: int, visible_height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Get rendered lines from the cache, or generate them."""
        bbox = self.bbox

        d_cols, d_aspect = self.datum.cell_size()
        d_rows = d_cols * d_aspect

        total_available_width = visible_width + bbox.left + bbox.right
        total_available_height = visible_height + bbox.top + bbox.bottom

        # Scale down the graphic to fit in the available space
        if d_rows > total_available_height or d_cols > total_available_width:
            if d_rows / total_available_height > d_cols / total_available_width:
                ratio = min(1, total_available_height / d_rows)
            else:
                ratio = min(1, total_available_width / d_cols)
        else:
            ratio = 1

        # Calculate the size and cropping bbox at which we want to display the graphic
        cols = floor(d_cols * ratio)
        rows = ceil(cols * d_aspect)
        d_bbox = DiInt(
            top=self.bbox.top,
            right=max(0, cols - (total_available_width - self.bbox.right)),
            bottom=max(0, rows - (total_available_height - self.bbox.bottom)),
            left=self.bbox.left,
        )

        def render_lines() -> list[StyleAndTextTuples]:
            """Render the lines to display in the control."""
            ft: list[StyleAndTextTuples] = []
            if visible_height >= 0:
                cmd = self.convert_data(BoundedWritePosition(0, 0, cols, rows, d_bbox))
                ft.extend(
                    split_lines(
                        to_formatted_text(
                            [
                                # Move cursor down and across by image height and width
                                (
                                    "",
                                    "\n".join(
                                        (visible_height) * [" " * (visible_width)]
                                    ),
                                ),
                                # Save position, then move back
                                ("[ZeroWidthEscape]", "\x1b[s"),
                                # Move cursor up if there is more than one line to display
                                *(
                                    [
                                        (
                                            "[ZeroWidthEscape]",
                                            f"\x1b[{visible_height - 1}A",
                                        )
                                    ]
                                    if visible_height > 1
                                    else []
                                ),
                                ("[ZeroWidthEscape]", f"\x1b[{visible_width}D"),
                                # Place the image without moving cursor
                                (
                                    "[ZeroWidthEscape]",
                                    passthrough(cmd, self.app.config),
                                ),
                                # ("[ZeroWidthEscape]", "XXXXX"),
                                # Restore the last known cursor position (at the bottom)
                                ("[ZeroWidthEscape]", "\x1b[u"),
                            ]
                        )
                    )
                )
            return ft

        key = (
            visible_width,
            self.app.color_palette,
            self.app.cell_size_px,
            self.bbox,
        )
        return self._format_cache.get(key, render_lines)


class ItermGraphicControl(GraphicControl):
    """A graphic control which displays images using iTerm's graphics protocol."""

    def convert_data(self, wp: WritePosition) -> str:
        """Convert the graphic's data to base64 data."""
        datum = self.datum
        bbox = wp.bbox if isinstance(wp, BoundedWritePosition) else DiInt(0, 0, 0, 0)
        # Crop image if necessary
        if any(bbox):
            import io

            image = datum.convert(to="pil", cols=wp.width, rows=wp.height)
            if image is not None:
                cell_size_x, cell_size_y = self.app.cell_size_px
                # Downscale image to fit target region for precise cropping
                image.thumbnail((wp.width * cell_size_x, wp.height * cell_size_y))
                left = bbox.left * cell_size_x
                top = bbox.top * cell_size_y
                right = (wp.width - bbox.right) * cell_size_x
                bottom = (wp.height - bbox.bottom) * cell_size_y
                upper, lower = sorted((top, bottom))
                image = image.crop((left, upper, right, lower))
                with io.BytesIO() as output:
                    image.save(output, format="PNG")
                    datum = Datum(data=output.getvalue(), format="png")

        if datum.format.startswith("base64-"):
            b64data = datum.data
        else:
            b64data = datum.convert(to="base64-png", cols=wp.width, rows=wp.height)
        return b64data.replace("\n", "").strip()

    def get_rendered_lines(
        self, visible_width: int, visible_height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Get rendered lines from the cache, or generate them."""
        bbox = self.bbox

        d_cols, d_aspect = self.datum.cell_size()
        d_rows = d_cols * d_aspect

        total_available_width = visible_width + bbox.left + bbox.right
        total_available_height = visible_height + bbox.top + bbox.bottom

        # Scale down the graphic to fit in the available space
        if d_rows > total_available_height or d_cols > total_available_width:
            if d_rows / total_available_height > d_cols / total_available_width:
                ratio = min(1, total_available_height / d_rows)
            else:
                ratio = min(1, total_available_width / d_cols)
        else:
            ratio = 1

        # Calculate the size and cropping bbox at which we want to display the graphic
        cols = floor(d_cols * ratio)
        rows = ceil(cols * d_aspect)
        d_bbox = DiInt(
            top=self.bbox.top,
            right=max(0, cols - (total_available_width - self.bbox.right)),
            bottom=max(0, rows - (total_available_height - self.bbox.bottom)),
            left=self.bbox.left,
        )

        def render_lines() -> list[StyleAndTextTuples]:
            """Render the lines to display in the control."""
            ft: list[StyleAndTextTuples] = []
            if (
                rows - d_bbox.top - d_bbox.bottom > 0
                and cols - d_bbox.left - d_bbox.right > 0
            ):
                b64data = self.convert_data(
                    BoundedWritePosition(0, 0, cols, rows, d_bbox)
                )
                cmd = f"\x1b]1337;File=inline=1;width={cols}:{b64data}\a"
                ft.extend(
                    split_lines(
                        to_formatted_text(
                            [
                                # Move cursor down and across by image height and width
                                (
                                    "",
                                    "\n".join(
                                        (visible_height) * [" " * (visible_width)]
                                    ),
                                ),
                                # Save position, then move back
                                ("[ZeroWidthEscape]", "\x1b[s"),
                                # Move cursor up if there is more than one line to display
                                *(
                                    [
                                        (
                                            "[ZeroWidthEscape]",
                                            f"\x1b[{visible_height - 1}A",
                                        )
                                    ]
                                    if visible_height > 1
                                    else []
                                ),
                                ("[ZeroWidthEscape]", f"\x1b[{visible_width}D"),
                                # Place the image without moving cursor
                                (
                                    "[ZeroWidthEscape]",
                                    passthrough(cmd, self.app.config),
                                ),
                                # Restore the last known cursor position (at the bottom)
                                ("[ZeroWidthEscape]", "\x1b[u"),
                            ]
                        )
                    )
                )
            return ft

        key = (
            visible_width,
            self.app.color_palette,
            self.app.cell_size_px,
            self.bbox,
        )
        return self._format_cache.get(key, render_lines)


class BaseKittyGraphicControl(GraphicControl):
    """Base graphic control with common methods for both styles of kitty display."""

    _kitty_image_count: ClassVar[int] = 1

    def __init__(
        self,
        datum: Datum,
        scale: float = 0,
        bbox: DiInt | None = None,
    ) -> None:
        """Create a new kitty graphic instance."""
        super().__init__(datum=datum, scale=scale, bbox=bbox)
        self.kitty_image_id = 0
        self.loaded = False
        self._datum_pad_cache: FastDictCache[tuple[Datum, int, int], Datum] = (
            FastDictCache(get_value=self._pad_datum, size=1)
        )

    def _pad_datum(self, datum: Datum, cell_size_x: int, cell_size_y: int) -> Datum:
        from PIL import ImageOps

        px, py = datum.pixel_size()

        if px and py:
            target_width = int((px + cell_size_x - 1) // cell_size_x * cell_size_x)
            target_height = int((py + cell_size_y - 1) // cell_size_y * cell_size_y)

            image = ImageOps.pad(
                datum.convert("pil").convert("RGBA"),
                (target_width, target_height),
                centering=(0, 0),
            )
            datum = Datum(
                image,
                format="pil",
                px=target_width,
                py=target_height,
                path=datum.path,
                align=datum.align,
            )
        return datum

    def convert_data(self, wp: WritePosition) -> str:
        """Convert the graphic's data to base64 data for kitty graphics protocol."""
        bbox = wp.bbox if isinstance(wp, BoundedWritePosition) else DiInt(0, 0, 0, 0)
        full_width = wp.width + bbox.left + bbox.right
        full_height = wp.height + bbox.top + bbox.bottom

        datum = self._datum_pad_cache[(self.datum, *self.app.cell_size_px)]
        return str(
            datum.convert(
                to="base64-png",
                cols=full_width,
                rows=full_height,
            )
        ).replace("\n", "")

    @staticmethod
    def _kitty_cmd(chunk: str = "", **params: Any) -> str:
        param_str = ",".join(
            [f"{key}={value}" for key, value in params.items() if value is not None]
        )
        cmd = f"\x1b_G{param_str}"
        if chunk:
            cmd += f";{chunk}"
        cmd += "\x1b\\"
        return cmd

    def load(self, rows: int, cols: int, bbox: DiInt) -> None:
        """Send the graphic to the terminal without displaying it."""
        data = self.convert_data(
            BoundedWritePosition(0, 0, width=cols, height=rows, bbox=bbox)
        )
        self.kitty_image_id = self._kitty_image_count
        self.__class__._kitty_image_count += 1

        while data:
            chunk, data = data[:4096], data[4096:]
            cmd = self._kitty_cmd(
                chunk=chunk,
                a="t",  # We are sending an image without displaying it
                t="d",  # Transferring the image directly
                i=self.kitty_image_id,  # Send a unique image number, wait for an image id
                # I=self.kitty_image_number,  # Send a unique image number, wait for an image id
                p=1,  # Placement ID
                q=2,  # No chatback
                f=100,  # Sending a PNG image
                C=1,  # Do not move the cursor
                m=1 if data else 0,  # Data will be chunked
            )
            self.app.output.write_raw(passthrough(cmd, self.app.config))
        self.app.output.flush()
        self.loaded = True

    def delete(self) -> None:
        """Delete the graphic from the terminal."""
        if self.kitty_image_id > 0:
            self.app.output.write_raw(
                passthrough(
                    self._kitty_cmd(
                        a="D",
                        d="I",
                        i=self.kitty_image_id,
                        q=2,
                    ),
                    self.app.config,
                )
            )
            self.app.output.flush()
            self.loaded = False

    def reset(self) -> None:
        """Hide and delete the kitty graphic from the terminal."""
        self.hide()
        self.delete()
        super().reset()

    def close(self) -> None:
        """Remove the displayed object entirely."""
        super().close()
        if not self.app.leave_graphics():
            self.delete()


class KittyGraphicControl(BaseKittyGraphicControl):
    """A graphic control which displays images using Kitty's graphics protocol."""

    def get_rendered_lines(
        self, visible_width: int, visible_height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Get rendered lines from the cache, or generate them."""
        bbox = self.bbox

        cell_size_px = self.app.cell_size_px
        datum = self._datum_pad_cache[(self.datum, *cell_size_px)]
        px, py = datum.pixel_size()
        # Fall back to a default pixel size
        px = px or 100
        py = py or 100

        d_cols, d_aspect = datum.cell_size()
        d_rows = d_cols * d_aspect

        total_available_width = visible_width + bbox.left + bbox.right
        total_available_height = visible_height + bbox.top + bbox.bottom

        # Scale down the graphic to fit in the available space
        if d_rows > total_available_height or d_cols > total_available_width:
            if d_rows / total_available_height > d_cols / total_available_width:
                ratio = min(1, total_available_height / d_rows)
            else:
                ratio = min(1, total_available_width / d_cols)
        else:
            ratio = 1

        # Calculate the size and cropping bbox at which we want to display the graphic
        cols = floor(d_cols * ratio)
        rows = ceil(cols * d_aspect)
        d_bbox = DiInt(
            top=self.bbox.top,
            right=max(0, cols - (total_available_width - self.bbox.right)),
            bottom=max(0, rows - (total_available_height - self.bbox.bottom)),
            left=self.bbox.left,
        )
        if not self.loaded:
            self.load(cols=cols, rows=rows, bbox=d_bbox)

        def render_lines() -> list[StyleAndTextTuples]:
            """Render the lines to display in the control."""
            ft: list[StyleAndTextTuples] = []
            if display_rows := rows - d_bbox.top - d_bbox.bottom:
                cmd = self._kitty_cmd(
                    a="p",  # Display a previously transmitted image
                    i=self.kitty_image_id,
                    p=1,  # Placement ID
                    m=0,  # No batches remaining
                    q=2,  # No backchat
                    c=cols - d_bbox.left - d_bbox.right,
                    r=display_rows,
                    C=1,  # 1 = Do move the cursor
                    # Horizontal pixel offset of the displayed image region
                    x=int(px * d_bbox.left / cols),
                    # Vertical pixel offset of the displayed image region
                    y=int(py * d_bbox.top / rows),
                    # Pixel width of the displayed image region
                    w=int(px * (cols - d_bbox.left - d_bbox.right) / cols),
                    # Pixel height of the displayed image region
                    h=int(py * (rows - d_bbox.top - d_bbox.bottom) / rows),
                )
                ft.extend(
                    split_lines(
                        to_formatted_text(
                            [
                                # Move cursor down and acoss by image height and width
                                (
                                    "",
                                    "\n".join(
                                        (visible_height) * [" " * (visible_width)]
                                    ),
                                ),
                                # Save position, then move back
                                ("[ZeroWidthEscape]", "\x1b[s"),
                                # Move cursor up if there is more than one line to display
                                *(
                                    [
                                        (
                                            "[ZeroWidthEscape]",
                                            f"\x1b[{visible_height - 1}A",
                                        )
                                    ]
                                    if visible_height > 1
                                    else []
                                ),
                                ("[ZeroWidthEscape]", f"\x1b[{visible_width}D"),
                                # Place the image without moving cursor
                                (
                                    "[ZeroWidthEscape]",
                                    passthrough(cmd, self.app.config),
                                ),
                                # Restore the last known cursor position (at the bottom)
                                ("[ZeroWidthEscape]", "\x1b[u"),
                            ]
                        )
                    )
                )
            else:
                ft.append([("[ZeroWidthEscape]", self.hide_cmd())])
            return ft

        key = (
            visible_width,
            self.app.color_palette,
            self.app.cell_size_px,
            self.bbox,
        )
        return self._format_cache.get(key, render_lines)

    def hide_cmd(self) -> str:
        """Generate a command to hide the graphic."""
        return passthrough(
            self._kitty_cmd(
                a="d",
                d="i",
                i=self.kitty_image_id,
                q=1,
            ),
            self.app.config,
        )

    def hide(self) -> None:
        """Hide the graphic from show without deleting it."""
        if self.kitty_image_id > 0:
            self.app.output.write_raw(self.hide_cmd())
            self.app.output.flush()


class KittyUnicodeGraphicControl(BaseKittyGraphicControl):
    """A graphic control which displays images using Kitty's Unicode placeholder mechanism."""

    PLACEHOLDER = "\U0010eeee"  # U+10EEEE placeholder character
    # fmt: off
    DIACRITICS = (  # Diacritics for encoding row/column numbers (0-9)
        "\u0305", "\u030d", "\u030e", "\u0310", "\u0312", "\u033d", "\u033e", "\u033f",
        "\u0346", "\u034a", "\u034b", "\u034c", "\u0350", "\u0351", "\u0352", "\u0357",
        "\u035b", "\u0363", "\u0364", "\u0365", "\u0366", "\u0367", "\u0368", "\u0369",
        "\u036a", "\u036b", "\u036c", "\u036d", "\u036e", "\u036f", "\u0483", "\u0484",
        "\u0485", "\u0486", "\u0487", "\u0592", "\u0593", "\u0594", "\u0595", "\u0597",
        "\u0598", "\u0599", "\u059c", "\u059d", "\u059e", "\u059f", "\u05a0", "\u05a1",
        "\u05a8", "\u05a9", "\u05ab", "\u05ac", "\u05af", "\u05c4", "\u0610", "\u0611",
        "\u0612", "\u0613", "\u0614", "\u0615", "\u0616", "\u0617", "\u0657", "\u0658",
        "\u0659", "\u065a", "\u065b", "\u065d", "\u065e", "\u06d6", "\u06d7", "\u06d8",
        "\u06d9", "\u06da", "\u06db", "\u06dc", "\u06df", "\u06e0", "\u06e1", "\u06e2",
        "\u06e4", "\u06e7", "\u06e8", "\u06eb", "\u06ec", "\u0730", "\u0732", "\u0733",
        "\u0735", "\u0736", "\u073a", "\u073d", "\u073f", "\u0740", "\u0741", "\u0743",
        "\u0745", "\u0747", "\u0749", "\u074a", "\u07eb", "\u07ec", "\u07ed", "\u07ee",
        "\u07ef", "\u07f0", "\u07f1", "\u07f3", "\u0816", "\u0817", "\u0818", "\u0819",
        "\u081b", "\u081c", "\u081d", "\u081e", "\u081f", "\u0820", "\u0821", "\u0822",
        "\u0823", "\u0825", "\u0826", "\u0827", "\u0829", "\u082a", "\u082b", "\u082c",
        "\u082d", "\u0951", "\u0953", "\u0954", "\u0f82", "\u0f83", "\u0f86", "\u0f87",
        "\u135d", "\u135e", "\u135f", "\u17dd", "\u193a", "\u1a17", "\u1a75", "\u1a76",
        "\u1a77", "\u1a78", "\u1a79", "\u1a7a", "\u1a7b", "\u1a7c", "\u1b6b", "\u1b6d",
        "\u1b6e", "\u1b6f", "\u1b70", "\u1b71", "\u1b72", "\u1b73", "\u1cd0", "\u1cd1",
        "\u1cd2", "\u1cda", "\u1cdb", "\u1ce0", "\u1dc0", "\u1dc1", "\u1dc3", "\u1dc4",
        "\u1dc5", "\u1dc6", "\u1dc7", "\u1dc8", "\u1dc9", "\u1dcb", "\u1dcc", "\u1dd1",
        "\u1dd2", "\u1dd3", "\u1dd4", "\u1dd5", "\u1dd6", "\u1dd7", "\u1dd8", "\u1dd9",
        "\u1dda", "\u1ddb", "\u1ddc", "\u1ddd", "\u1dde", "\u1ddf", "\u1de0", "\u1de1",
        "\u1de2", "\u1de3", "\u1de4", "\u1de5", "\u1de6", "\u1dfe", "\u20d0", "\u20d1",
        "\u20d4", "\u20d5", "\u20d6", "\u20d7", "\u20db", "\u20dc", "\u20e1", "\u20e7",
        "\u20e9", "\u20f0", "\u2cef", "\u2cf0", "\u2cf1", "\u2de0", "\u2de1", "\u2de2",
        "\u2de3", "\u2de4", "\u2de5", "\u2de6", "\u2de7", "\u2de8", "\u2de9", "\u2dea",
        "\u2deb", "\u2dec", "\u2ded", "\u2dee", "\u2def", "\u2df0", "\u2df1", "\u2df2",
        "\u2df3", "\u2df4", "\u2df5", "\u2df6", "\u2df7", "\u2df8", "\u2df9", "\u2dfa",
        "\u2dfb", "\u2dfc", "\u2dfd", "\u2dfe", "\u2dff", "\ua66f", "\ua67c", "\ua67d",
        "\ua6f0", "\ua6f1", "\ua8e0", "\ua8e1", "\ua8e2", "\ua8e3", "\ua8e4", "\ua8e5",
        "\ua8e6", "\ua8e7", "\ua8e8", "\ua8e9", "\ua8ea", "\ua8eb", "\ua8ec", "\ua8ed",
        "\ua8ee", "\ua8ef", "\ua8f0", "\ua8f1", "\uaab0", "\uaab2", "\uaab3", "\uaab7",
        "\uaab8", "\uaabe", "\uaabf", "\uaac1", "\ufe20", "\ufe21", "\ufe22", "\ufe23",
        "\ufe24", "\ufe25", "\ufe26",
        "\U00010a0f", "\U00010a38", "\U0001d185", "\U0001d186", "\U0001d187",
        "\U0001d188", "\U0001d189", "\U0001d1aa", "\U0001d1ab", "\U0001d1ac",
        "\U0001d1ad", "\U0001d242", "\U0001d243", "\U0001d244",
    )
    # fmt: on

    def __init__(
        self,
        datum: Datum,
        scale: float = 0,
        bbox: DiInt | None = None,
    ) -> None:
        """Create a new kitty graphic instance."""
        super().__init__(datum, scale, bbox)
        self.placements: set[tuple[int, int]] = set()

    def get_rendered_lines(
        self, visible_width: int, visible_height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Get rendered lines from the cache, or generate them."""
        bbox = self.bbox

        cell_size_px = self.app.cell_size_px
        datum = self._datum_pad_cache[(self.datum, *cell_size_px)]
        px, py = datum.pixel_size()
        # Fall back to a default pixel size
        px = px or 100
        py = py or 100

        d_cols, d_aspect = datum.cell_size()
        d_rows = d_cols * d_aspect

        total_available_width = visible_width + bbox.left + bbox.right
        total_available_height = visible_height + bbox.top + bbox.bottom

        # Scale down the graphic to fit in the available space
        if d_rows > total_available_height or d_cols > total_available_width:
            if d_rows / total_available_height > d_cols / total_available_width:
                ratio = min(1, total_available_height / d_rows)
            else:
                ratio = min(1, total_available_width / d_cols)
        else:
            ratio = 1

        # Calculate the size and cropping bbox at which we want to display the graphic
        cols = floor(d_cols * ratio)
        rows = ceil(cols * d_aspect)
        d_bbox = DiInt(
            top=self.bbox.top,
            right=max(0, cols - (total_available_width - self.bbox.right)),
            bottom=max(0, rows - (total_available_height - self.bbox.bottom)),
            left=self.bbox.left,
        )
        if not self.loaded:
            self.load(cols=cols, rows=rows, bbox=DiInt(0, 0, 0, 0))

        # Add virtual placement for this size if required
        if (cols, rows) not in self.placements:
            cmd = self._kitty_cmd(
                a="p",  # Display a previously transmitted image
                i=self.kitty_image_id,
                p=1,  # Placement ID
                U=1,  # Create a virtual placement
                c=cols,
                r=rows,
                q=2,
            )
            self.app.output.write_raw(passthrough(cmd, self.app.config))
            self.app.output.flush()
            self.placements.add((cols, rows))

        def render_lines() -> list[StyleAndTextTuples]:
            """Render the lines to display in the control."""
            ft: StyleAndTextTuples = []

            # Generate placeholder grid
            row_start = d_bbox.top
            row_stop = rows - d_bbox.bottom
            col_start = d_bbox.left
            col_stop = cols - d_bbox.right
            placeholder = self.PLACEHOLDER
            diacritics = self.DIACRITICS
            for row in range(row_start, row_stop):
                for col in range(col_start, col_stop):
                    ft.extend(
                        [
                            # We set the ptk-color for the last column so the renderer
                            # knows to change the color back after this gets rendered.
                            ("fg:#888" if col == col_stop - 1 else "", " "),
                            (
                                "[ZeroWidthEscape]",
                                # We move the cursor back a cell before writing the
                                # kitty unicode char using a ZWE
                                "\b"
                                # Set the kitty graphic and placement we want to render
                                # by manually setting an 8-bit foregroun color.
                                # The placement ID is set to 1 using underline color.
                                f"\x1b[38;5;{self.kitty_image_id}m\x1b[58;1m"
                                # Writing the unicode char moves the cursor forward
                                # again to where the renderer expects it to be
                                f"{placeholder}{diacritics[row]}{diacritics[col]}",
                            ),
                        ]
                    )
                ft.append(("", "\n"))
            return list(split_lines(ft))

        key = (
            visible_width,
            self.app.color_palette,
            self.app.cell_size_px,
            bbox,
        )
        return self._format_cache.get(key, render_lines)


class NotVisible(Exception):
    """Exception to signal that a graphic is not currently visible."""


class GraphicWindow(Window):
    """A window responsible for displaying terminal graphics content.

    The content is displayed floating on top of a target window.

    The graphic will be displayed if:
    - a completion menu is not being shown
    - a dialog is not being shown
    - a menu is not being shown
    - the output it attached to is fully in view
    """

    content: GraphicControl

    def __init__(
        self,
        content: GraphicControl,
        get_position: Callable[[Screen], BoundedWritePosition],
        filter: FilterOrBool = True,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initiate a new :py:class:`GraphicWindow` object.

        Args:
            content: A control which generates the graphical content to display
            get_position: A function which returns the position of the graphic
            filter: A filter which determines if the graphic should be shown
            args: Positional arguments for :py:method:`Window.__init__`
            kwargs: Key-word arguments for :py:method:`Window.__init__`
        """
        super().__init__(*args, **kwargs)
        self.content = content
        self.get_position = get_position
        self.filter = ~has_float & to_filter(filter)
        self._pre_rendered = False

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Draw the graphic window's contents to the screen if required."""
        # Pre-convert datum for this write position so result is cached
        if not self._pre_rendered:
            self.content.convert_data(write_position)
            self._pre_rendered = True
        filter_value = self.filter()
        if filter_value:
            try:
                new_write_position = self.get_position(screen)
            except NotVisible:
                pass
            else:
                # Inform the graphic control of the calculated cropping region
                self.content.bbox = new_write_position.bbox

                # Draw the graphic content to the screen
                if (
                    new_write_position
                    and new_write_position.width
                    and new_write_position.height
                ):
                    # Do not pass the bbox on to the window when writing
                    new_write_position.bbox = DiInt(0, 0, 0, 0)
                    super().write_to_screen(
                        screen,
                        MouseHandlers(),  # Do not let the float add mouse events
                        new_write_position,
                        # Force renderer refreshes by constantly changing the style
                        f"{parent_style} class:render-{get_app().render_counter}",
                        erase_bg=True,
                        z_index=z_index,
                    )
                    return

        # Otherwise hide the content (required for kitty graphics)
        if not get_app().leave_graphics():
            self.content.hide()

    def _fill_bg(
        self,
        screen: Screen,
        # mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        erase_bg: bool,
    ) -> None:
        """Erase/fill the background."""
        char: str | None = self.char() if callable(self.char) else self.char
        if erase_bg or char:
            wp = write_position
            char_obj = Char(char or " ", f"class:render-{get_app().render_counter}")

            # mouse_handlers_dict = mouse_handlers.mouse_handlers
            for y in range(wp.ypos, wp.ypos + wp.height):
                row = screen.data_buffer[y]
                # mouse_handler_row = mouse_handlers_dict[y]
                for x in range(wp.xpos, wp.xpos + wp.width):
                    row[x] = char_obj
                    # mouse_handler_row[x] = lambda e: NotImplemented


def select_graphic_control(format_: str) -> type[GraphicControl] | None:
    """Determine which graphic control to use."""
    SelectedGraphicControl: type[GraphicControl] | None = None
    app = get_app()
    preferred_graphics_protocol = app.config.graphics
    useable_graphics_controls: list[type[GraphicControl]] = []
    _in_mplex = in_mplex()
    force_graphics = app.config.force_graphics

    if preferred_graphics_protocol != "none":
        if (app.term_graphics_iterm or force_graphics) and find_route(
            format_, "base64-png"
        ):
            useable_graphics_controls.append(ItermGraphicControl)
        if (
            preferred_graphics_protocol == "iterm"
            and ItermGraphicControl in useable_graphics_controls
            # Iterm does not work in mplex without pass-through
            and (not _in_mplex or (_in_mplex and force_graphics))
        ):
            SelectedGraphicControl = ItermGraphicControl
        elif (
            (app.term_graphics_kitty or force_graphics)
            and find_route(format_, "base64-png")
            # Kitty does not work in mplex without pass-through
            and (not _in_mplex or (_in_mplex and force_graphics))
        ):
            useable_graphics_controls.append(KittyGraphicControl)
            useable_graphics_controls.append(KittyUnicodeGraphicControl)
        if (
            preferred_graphics_protocol == "kitty"
            and KittyGraphicControl in useable_graphics_controls
        ):
            SelectedGraphicControl = KittyGraphicControl
        elif (
            preferred_graphics_protocol == "kitty-unicode"
            and KittyUnicodeGraphicControl in useable_graphics_controls
        ):
            SelectedGraphicControl = KittyUnicodeGraphicControl
        # Tmux now supports sixels (>=3.4)
        elif (app.term_graphics_sixel or force_graphics) and find_route(
            format_, "sixel"
        ):
            useable_graphics_controls.append(SixelGraphicControl)
        if (
            preferred_graphics_protocol == "sixel"
            and SixelGraphicControl in useable_graphics_controls
        ):
            SelectedGraphicControl = SixelGraphicControl

        if SelectedGraphicControl is None and useable_graphics_controls:
            SelectedGraphicControl = useable_graphics_controls[0]

    return SelectedGraphicControl


class GraphicProcessor:
    """Class which loads and positions graphics references in a :py:class:`UIContent`."""

    def __init__(self, control: UIControl) -> None:
        """Initialize a new graphic processor."""
        self.control = control

        self.positions: dict[str, Point] = {}
        self._position_cache: FastDictCache[tuple[UIContent], dict[str, Point]] = (
            FastDictCache(self._load_positions, size=1_000)
        )
        self._float_cache: FastDictCache[tuple[str], Float | None] = FastDictCache(
            self.get_graphic_float, size=1_000
        )
        self.app = get_app()

    def load(self, content: UIContent) -> None:
        """Check for graphics in lines of text."""
        self.positions = self._position_cache[content,]

    def _load_positions(self, content: UIContent) -> dict[str, Point]:
        positions = {}
        get_line = content.get_line
        for y in range(content.line_count):
            line = get_line(y)
            x = 0
            for style, text, *_ in line:
                for part in style.split():
                    if part.startswith("[Graphic_"):
                        key = part[9:-1]
                        positions[key] = Point(x, y)
                        # Get graphic float for this image and update its position
                        graphic_float = self._float_cache[key,]
                        # Register graphic with application
                        if graphic_float:
                            self.app.graphics.add(graphic_float)
                    if part in _ZERO_WIDTH_FRAGMENTS:
                        break
                else:
                    x += get_cwidth(text)
        return positions

    def _get_position(
        self, key: str, rows: int, cols: int
    ) -> Callable[[Screen], BoundedWritePosition]:
        """Return a function that returns the current bounded graphic position."""

        def get_graphic_position(screen: Screen) -> BoundedWritePosition:
            """Get the position and bbox of a graphic."""
            if key not in self.positions:
                raise NotVisible

            # Find the control's window
            window: Window | None = None
            for _window in get_app().layout.find_all_windows():
                if _window.content == self.control:
                    window = _window
                    break

            if window not in screen.visible_windows:
                raise NotVisible

            render_info = window.render_info
            if render_info is None:
                raise NotVisible

            # Hide graphic if control is not in layout
            win_wp = screen.visible_windows_to_write_positions.get(window)
            if win_wp is None:
                raise NotVisible

            if isinstance(win_wp, BoundedWritePosition):
                win_bbox = win_wp.bbox
            else:
                win_bbox = DiInt(0, 0, 0, 0)

            render_info = window.render_info
            if render_info is None:
                raise NotVisible

            x, y = self.positions[key]
            win_content_height = render_info.ui_content.line_count
            win_content_width = win_wp.width  # TODO - get the actual content width

            horizontal_scroll = getattr(render_info, "horizontal_scroll", 0)
            vertical_scroll = render_info.vertical_scroll
            if (
                horizontal_scroll >= x + win_content_width
                or vertical_scroll >= y + win_content_height
            ):
                raise NotVisible

            bbox = DiInt(
                top=max(0, win_bbox.top - y + vertical_scroll),
                right=win_bbox.right,
                bottom=max(
                    0, win_bbox.bottom - (win_wp.height - y - rows) - vertical_scroll
                ),
                left=win_bbox.left,
            )

            xpos = win_wp.xpos + x + win_bbox.left
            ypos = max(
                win_wp.ypos + win_bbox.top,
                win_wp.ypos + y + max(0, win_bbox.top - y) - min(y, vertical_scroll),
            )
            width = max(0, cols - bbox.left - bbox.right)
            height = max(0, rows - bbox.top - bbox.bottom)

            return BoundedWritePosition(
                xpos=xpos, ypos=ypos, width=width, height=height, bbox=bbox
            )

        return get_graphic_position

    def get_graphic_float(self, key: str) -> Float | None:
        """Create a graphical float for an image."""
        sized_datum = Datum.get_size(key)
        if sized_datum is None:
            log.debug("Datum not found for graphic '%s'", key)
            return None

        datum, (rows, cols) = sized_datum

        GraphicControl = select_graphic_control(format_=datum.format)
        if GraphicControl is None:
            # log.debug("Terminal graphics not supported or format not graphical")
            return None

        bg_color = datum.bg
        graphic_float = Float(
            graphic_window := GraphicWindow(
                content=(graphic_control := GraphicControl(datum)),
                get_position=self._get_position(key, rows, cols),
                style=f"bg:{bg_color}" if bg_color else "",
                align=datum.align,
            )
        )
        # Register graphic with application
        (app_graphics := self.app.graphics).add(graphic_float)
        # Hide the graphic from app if the float is deleted
        weak_float_ref = weakref.ref(graphic_float)
        graphic_window.filter &= Condition(lambda: weak_float_ref() in app_graphics)
        # Hide the graphic from terminal if the float is deleted
        weakref.finalize(graphic_float, graphic_control.close)

        return graphic_float



"""A text base user interface for euporie."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

from prompt_toolkit.application.current import get_app as ptk_get_app
from prompt_toolkit.filters.app import (
    has_completions,
    is_done,
    is_searching,
    renderer_height_is_known,
)
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Float,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.dimension import Dimension

from euporie.console.tabs.console import Console
from euporie.core.app.app import BaseApp
from euporie.core.bars.command import CommandBar
from euporie.core.bars.menu import ToolbarCompletionsMenu
from euporie.core.bars.search import SearchBar
from euporie.core.bars.status import StatusBar
from euporie.core.filters import has_dialog
from euporie.core.layout.mouse import DisableMouseOnScroll
from euporie.core.widgets.dialog import (
    AboutDialog,
    ConfirmDialog,
    NoKernelsDialog,
    SaveAsDialog,
    SelectKernelDialog,
    ShortcutsDialog,
)
from euporie.core.widgets.logo import Logo
from euporie.core.widgets.pager import Pager
from euporie.core.widgets.palette import CommandPalette

if TYPE_CHECKING:
    from typing import Any, TypeVar

    from prompt_toolkit.application.application import Application

    _AppResult = TypeVar("_AppResult")

log = logging.getLogger(__name__)


def get_app() -> ConsoleApp:
    """Get the current application."""
    return cast("ConsoleApp", ptk_get_app())


class ConsoleApp(BaseApp):
    """Conole app.

    An interactive console which connects to Jupyter kernels and displays
    rich output in the terminal.

    """

    name = "console"

    def __init__(self, **kwargs: Any) -> None:
        """Create a new euporie text user interface application instance."""
        # Set default application options
        kwargs.setdefault("extend_renderer_height", True)
        kwargs.setdefault("title", "euporie-console")
        kwargs.setdefault("full_screen", False)
        kwargs.setdefault("leave_graphics", True)
        kwargs.setdefault("mouse_support", self.config.filters.mouse_support)

        # Initialize the application
        super().__init__(**kwargs)

        self.bindings_to_load += ["euporie.console.app.ConsoleApp"]

        self.tabs = []

    def pre_run(self, app: Application | None = None) -> None:
        """Continue loading the app."""
        super().pre_run(app)
        # Add a toolbar completion menu
        self.menus["toolbar_completions"] = Float(
            content=ToolbarCompletionsMenu(), ycursor=True, transparent=True
        )

    def _get_reserved_height(self) -> Dimension:
        if has_dialog():
            return Dimension(min=15)
        elif has_completions():
            return Dimension(min=5)
        else:
            return Dimension(min=1)

    def load_container(self) -> FloatContainer:
        """Return a container with all opened tabs."""
        self.tabs = [Console(self)]

        self.command_bar = CommandBar()
        self.search_bar = SearchBar()
        self.pager = Pager()

        self.dialogs["command-palette"] = CommandPalette(self)
        self.dialogs["about"] = AboutDialog(self)
        self.dialogs["save-as"] = SaveAsDialog(self)
        self.dialogs["no-kernels"] = NoKernelsDialog(self)
        self.dialogs["change-kernel"] = SelectKernelDialog(self)
        self.dialogs["shortcuts"] = ShortcutsDialog(self)
        self.dialogs["confirm"] = ConfirmDialog(self)

        return FloatContainer(
            DisableMouseOnScroll(
                HSplit(
                    [
                        self.tabs[0],
                        ConditionalContainer(
                            HSplit(
                                [
                                    # Fill empty space below input
                                    Window(
                                        height=self._get_reserved_height,
                                        style="class:default",
                                    ),
                                    self.pager,
                                    ConditionalContainer(
                                        VSplit(
                                            [
                                                Logo(),
                                                self.command_bar,
                                                self.search_bar,
                                                StatusBar(),
                                            ]
                                        ),
                                        filter=~is_searching,
                                    ),
                                ],
                            ),
                            filter=~self.redrawing
                            & ~is_done
                            & renderer_height_is_known,
                        ),
                    ]
                )
            ),
            floats=self.floats,  # type: ignore
        )

    def exit(
        self,
        result: _AppResult | None = None,
        exception: BaseException | type[BaseException] | None = None,
        style: str = "",
    ) -> None:
        """Close all tabs on exit."""
        for tab in self.tabs:
            tab.close()

        if result is not None:
            super().exit(result=result, style=style)
        elif exception is not None:
            super().exit(exception=exception, style=style)
        else:
            super().exit()

    # ################################# Key Bindings ##################################

    # register_bindings({"euporie.console.app.ConsoleApp": {}})



"""Define a status-bar widget."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.layout import containers
from prompt_toolkit.layout.containers import ConditionalContainer, WindowAlign
from prompt_toolkit.layout.containers import (
    to_container as ptk_to_container,
)
from prompt_toolkit.layout.controls import FormattedTextControl

from euporie.core.app.current import get_app
from euporie.core.filters import has_toolbar
from euporie.core.layout.containers import VSplit, Window

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Callable

    from prompt_toolkit.filters.base import FilterOrBool
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.formatted_text.base import AnyFormattedText
    from prompt_toolkit.layout.containers import AnyContainer, Container

    StatusBarFields = tuple[Sequence[AnyFormattedText], Sequence[AnyFormattedText]]

log = logging.getLogger(__name__)


_CONTAINER_STATUSES: WeakKeyDictionary[
    Container, Callable[[], StatusBarFields | None]
] = WeakKeyDictionary()


def to_container(container: AnyContainer) -> Container:
    """Monkey-patch `to_container` to collect container status functions."""
    result = ptk_to_container(container)
    if hasattr(container, "__pt_status__"):
        _CONTAINER_STATUSES[result] = container.__pt_status__
    return result


containers.to_container = to_container


class StatusContainer:
    """A container which allows attaching a status function."""

    def __init__(
        self, body: AnyContainer, status: Callable[[], StatusBarFields | None]
    ) -> None:
        """Initiate a new instance with a body and status function."""
        self.body = body
        self.status = status

    def __pt_status__(self) -> StatusBarFields | None:
        """Return the status fields."""
        return self.status()

    def __pt_container__(self) -> AnyContainer:
        """Return the body container."""
        return self.body


class StatusBar:
    """A status bar which shows the status of the current tab."""

    def __init__(
        self, extra_filter: FilterOrBool = True, default: StatusBarFields | None = None
    ) -> None:
        """Create a new status bar instance."""
        self.default: StatusBarFields = default or ([], [])
        self._status_cache: FastDictCache[tuple[int], list[StyleAndTextTuples]] = (
            FastDictCache(self._status, size=1)
        )
        self.container = ConditionalContainer(
            content=VSplit(
                [
                    Window(
                        FormattedTextControl(
                            lambda: self._status_cache[get_app().render_counter,][0]
                        ),
                        style="class:status",
                    ),
                    Window(
                        FormattedTextControl(
                            lambda: self._status_cache[get_app().render_counter,][1]
                        ),
                        style="class:status.right",
                        align=WindowAlign.RIGHT,
                    ),
                ],
                height=1,
            ),
            filter=get_app().config.filters.show_status_bar
            & ~has_toolbar
            & to_filter(extra_filter),
        )

    def _status(self, render_counter: int = 0) -> list[StyleAndTextTuples]:
        """Load and format the current status bar entries."""
        layout = get_app().layout

        # entries = self.default
        entries: tuple[list[AnyFormattedText], list[AnyFormattedText]] = ([], [])

        current: Container | None = layout.current_window
        while current:
            if (
                callable(
                    func := (
                        _CONTAINER_STATUSES.get(current)
                        or getattr(current, "__pt_status__", None)
                    )
                )
                and (result := func()) is not None
            ):
                # Add parent entries to start of left side
                entries[0][0:0] = result[0]
                # Add parent entries to end of right side
                entries[1].extend(result[1])
            current = layout._child_to_parent.get(current)

        # Format the status entries
        output: list[StyleAndTextTuples] = []
        for entry in entries:
            output.append([])
            for field in entry:
                if field:
                    output[-1] += [
                        ("class:status-sep", "▌"),
                        *to_formatted_text(field, style="class:status-field"),
                        ("class:status-sep", "▐"),
                    ]
        return output

    def __pt_container__(self) -> AnyContainer:
        """Return the widget's container."""
        return self.container



"""Contains commands for tabs."""

from __future__ import annotations

import logging

from euporie.core.commands import add_cmd
from euporie.core.filters import tab_can_save, tab_has_focus

log = logging.getLogger(__name__)


@add_cmd(filter=tab_can_save, aliases=["w"])
def _save_file(path: str = "") -> None:
    """Save the current file."""
    from euporie.core.app.current import get_app

    if (tab := get_app().tab) is not None:
        from upath import UPath

        try:
            tab._save(UPath(path) if path else None)
        except NotImplementedError:
            pass


@add_cmd(aliases=["wq", "x"])
def _save_and_quit(path: str = "") -> None:
    """Save the current tab then quits euporie."""
    from euporie.core.app.current import get_app

    app = get_app()
    if (tab := app.tab) is not None:
        from upath import UPath

        try:
            tab.save(UPath(path) if path else None)
        except NotImplementedError:
            pass

    app.exit()


@add_cmd(
    menu_title="Save As…",
    filter=tab_can_save,
)
def _save_as(path: str = "") -> None:
    """Save the current file at a new location."""
    if path:
        _save_file(path)
    else:
        from euporie.core.app.current import get_app

        app = get_app()
        if dialog := app.dialogs.get("save-as"):
            dialog.show(tab=app.tab)


@add_cmd(filter=tab_has_focus, title="Refresh the current tab")
def _refresh_tab() -> None:
    """Reload the tab contents and reset the tab."""
    from euporie.core.app.current import get_app

    if (tab := get_app().tab) is not None:
        tab.reset()


# Depreciated v2.5.0
@add_cmd(filter=tab_has_focus, title="Reset the current tab")
def _reset_tab() -> None:
    log.warning(
        "The `reset-tab` command was been renamed to `refresh-tab` in v2.5.0,"
        " and will be removed in a future version"
    )
    _refresh_tab()



"""Define a dummy application."""

from euporie.core.app.app import BaseApp
from euporie.core.layout.containers import DummyContainer


class DummyApp(BaseApp):
    """An empty application which does nothing."""

    def load_container(self) -> DummyContainer:
        """Load a dummy container as the root container."""
        return DummyContainer()
